/*
 * TLB exception handling code for r4k.
 *
 * Copyright (C) 1994, 1995, 1996 by Ralf Baechle and Andreas Busse
 *
 * Multi-cpu abstraction and reworking:
 * Copyright (C) 1996 David S. Miller (dm@engr.sgi.com)
 *
 * Carsten Langgaard, carstenl@mips.com
 * Copyright (C) 2000 MIPS Technologies, Inc.  All rights reserved.
 */
#include <linux/config.h>
#include <linux/init.h>

#include <asm/asm.h>
#include <asm/current.h>

#ifndef CONFIG_KBUILD_2_5
#include <asm/offset.h>
#else
#include "asm-offsets.h"
#endif

#include <asm/cachectl.h>
#include <asm/fpregdef.h>
#include <asm/mipsregs.h>
#include <asm/page.h>
#include <asm/pgtable-bits.h>
#include <asm/processor.h>
#include <asm/regdef.h>
#include <asm/stackframe.h>

#define TLB_OPTIMIZE /* If you are paranoid, disable this. */

#ifdef CONFIG_64BIT_PHYS_ADDR
#define PTE_L		ld
#define PTE_S		sd
#define PTE_SRL		dsrl
#define P_MTC0		dmtc0
#define PTE_SIZE	8
#define PTEP_INDX_MSK	0xff0
#define PTE_INDX_MSK	0xff8
#define PTE_INDX_SHIFT	9
#else
#define PTE_L		lw
#define PTE_S		sw
#define PTE_SRL		srl
#define P_MTC0		mtc0
#define PTE_SIZE	4
#define PTEP_INDX_MSK	0xff8
#define PTE_INDX_MSK	0xffc
#define PTE_INDX_SHIFT	10
#endif

#ifdef CONFIG_TIVO_DEVEL
#define CONFIG_DEBUG_MEMPOOL
#endif
	__INIT

#ifdef CONFIG_64BIT_PHYS_ADDR
#define GET_PTE_OFF(reg)
#elif CONFIG_CPU_VR41XX
#define GET_PTE_OFF(reg)	srl	reg, reg, 3
#else
#define GET_PTE_OFF(reg)	srl	reg, reg, 1
#endif

/*
 * These handlers much be written in a relocatable manner
 * because based upon the cpu type an arbitrary one of the
 * following pieces of code will be copied to the KSEG0
 * vector location.
 */
	/* TLB refill, EXL == 0, R4xx0, non-R4600 version */
	.set	noreorder
	.set	noat
	LEAF(except_vec0_r4000)
	.set	mips3
#ifdef CONFIG_DEBUG_MEMPOOL
	mfc0	k0, CP0_BADVADDR		# Get faulting address
	srl	k1, k0, 28
	xori	k1, 0x4
	beqz	k1, 4f
skip_mempool_check:
#endif // CONFIG_DEBUG_MEMPOOL
#ifdef CONFIG_SMP
	 mfc0	k1, CP0_CONTEXT
	la	k0, pgd_current
	srl	k1, 23
	sll	k1, 2				# log2(sizeof(pgd_t)
	addu	k1, k0, k1
	lw	k1, (k1)
#ifdef CONFIG_DEBUG_MEMPOOL
	mfc0	k0, CP0_BADVADDR		# Get faulting addres
#endif // CONFIG_DEBUG_MEMPOOL
#else // CONFIG_SMP
	 lw	k1, pgd_current			# get pgd pointer
#endif // else CONFIG_SMP
#ifndef CONFIG_DEBUG_MEMOOL
	mfc0	k0, CP0_BADVADDR		# Get faulting address
#endif
	srl	k0, k0, _PGDIR_SHIFT		# get pgd only bits

	sll	k0, k0, 2
	addu	k1, k1, k0			# add in pgd offset
	mfc0	k0, CP0_CONTEXT			# get context reg
	lw	k1, (k1)
	GET_PTE_OFF(k0)				# get pte offset
	and	k0, k0, PTEP_INDX_MSK
	addu	k1, k1, k0			# add in offset
	PTE_L	k0, 0(k1)			# get even pte
	PTE_L	k1, PTE_SIZE(k1)		# get odd pte
#ifdef CONFIG_BMIPS3300_WAR
	bltz	k0, 2f
#endif
	PTE_SRL	k0, k0, 6			# convert to entrylo0
	P_MTC0	k0, CP0_ENTRYLO0		# load it
#ifdef CONFIG_BMIPS3300_WAR
	bltz	k1, 3f
#endif
	PTE_SRL	k1, k1, 6			# convert to entrylo1
	P_MTC0	k1, CP0_ENTRYLO1		# load it
#ifndef CONFIG_BMIPS3300_WAR
	b	1f
#endif
	tlbwr					# write random tlb entry
1:
	nop
	eret					# return from trap
#ifdef CONFIG_BMIPS3300_WAR
2:
	j	fixup_even_page
	 sll	k0, 7
3:
	j	fixup_odd_page
	 sll	k1, 7
#endif
#ifdef CONFIG_DEBUG_MEMPOOL
4:	j	debug_mempool_access
	 mfc0	k0, CP0_EPC
#endif

	END(except_vec0_r4000)

	/* TLB refill, EXL == 0, R4600 version */
	LEAF(except_vec0_r4600)
	.set	mips3
	mfc0	k0, CP0_BADVADDR
	srl	k0, k0, _PGDIR_SHIFT
	lw	k1, pgd_current			# get pgd pointer
	sll	k0, k0, 2			# log2(sizeof(pgd_t)
	addu	k1, k1, k0
	mfc0	k0, CP0_CONTEXT
	lw	k1, (k1)
#ifndef CONFIG_64BIT_PHYS_ADDR
	srl	k0, k0, 1
#endif
	and	k0, k0, PTEP_INDX_MSK
	addu	k1, k1, k0
	PTE_L	k0, 0(k1)
	PTE_L	k1, PTE_SIZE(k1)
	PTE_SRL	k0, k0, 6
	P_MTC0	k0, CP0_ENTRYLO0
	PTE_SRL	k1, k1, 6
	P_MTC0	k1, CP0_ENTRYLO1
	nop
	tlbwr
	nop
	eret
	END(except_vec0_r4600)

	/* TLB refill, EXL == 0, R52x0 "Nevada" version */
        /*
         * This version has a bug workaround for the Nevada.  It seems
         * as if under certain circumstances the move from cp0_context
         * might produce a bogus result when the mfc0 instruction and
         * it's consumer are in a different cacheline or a load instruction,
         * probably any memory reference, is between them.  This is
         * potencially slower than the R4000 version, so we use this
         * special version.
         */
	.set	noreorder
	.set	noat
	LEAF(except_vec0_nevada)
	.set	mips3
	mfc0	k0, CP0_BADVADDR		# Get faulting address
	srl	k0, k0, _PGDIR_SHIFT		# get pgd only bits
	lw	k1, pgd_current			# get pgd pointer
	sll	k0, k0, 2			# log2(sizeof(pgd_t)
	addu	k1, k1, k0			# add in pgd offset
	lw	k1, (k1)
	mfc0	k0, CP0_CONTEXT			# get context reg
#ifndef CONFIG_64BIT_PHYS_ADDR
	srl	k0, k0, 1			# get pte offset
#endif
	and	k0, k0, PTEP_INDX_MSK
	addu	k1, k1, k0			# add in offset
	PTE_L	k0, 0(k1)			# get even pte
	PTE_L	k1, PTE_SIZE(k1)		# get odd pte
	PTE_SRL	k0, k0, 6			# convert to entrylo0
	P_MTC0	k0, CP0_ENTRYLO0		# load it
	PTE_SRL	k1, k1, 6			# convert to entrylo1
	P_MTC0	k1, CP0_ENTRYLO1		# load it
	nop					# QED specified nops
	nop
	tlbwr					# write random tlb entry
	nop					# traditional nop
	eret					# return from trap
	END(except_vec0_nevada)

	/* TLB refill, EXL == 0, R4[40]00/R5000 badvaddr hwbug version */
	LEAF(except_vec0_r45k_bvahwbug)
	.set	mips3
	mfc0	k0, CP0_BADVADDR
	srl	k0, k0, _PGDIR_SHIFT
	lw	k1, pgd_current			# get pgd pointer
	sll	k0, k0, 2			# log2(sizeof(pgd_t)
	addu	k1, k1, k0
	mfc0	k0, CP0_CONTEXT
	lw	k1, (k1)
#ifndef CONFIG_64BIT_PHYS_ADDR
	srl	k0, k0, 1
#endif
	and	k0, k0, PTEP_INDX_MSK
	addu	k1, k1, k0
	PTE_L	k0, 0(k1)
	PTE_L	k1, PTE_SIZE(k1)
	nop				/* XXX */
	tlbp
	PTE_SRL	k0, k0, 6
	P_MTC0	k0, CP0_ENTRYLO0
	PTE_SRL	k1, k1, 6
	mfc0	k0, CP0_INDEX
	P_MTC0	k1, CP0_ENTRYLO1
	bltzl	k0, 1f
	tlbwr
1:
	nop
	eret
	END(except_vec0_r45k_bvahwbug)

#ifdef CONFIG_SMP
	/* TLB refill, EXL == 0, R4000 MP badvaddr hwbug version */
	LEAF(except_vec0_r4k_mphwbug)
	.set	mips3
	mfc0	k0, CP0_BADVADDR
	srl	k0, k0, _PGDIR_SHIFT
	lw	k1, pgd_current			# get pgd pointer
	sll	k0, k0, 2			# log2(sizeof(pgd_t)
	addu	k1, k1, k0
	mfc0	k0, CP0_CONTEXT
	lw	k1, (k1)
#ifndef CONFIG_64BIT_PHYS_ADDR
	srl	k0, k0, 1
#endif
	and	k0, k0, PTEP_INDX_MSK
	addu	k1, k1, k0
	PTE_L	k0, 0(k1)
	PTE_L	k1, PTE_SIZE(k1)
	nop				/* XXX */
	tlbp
	PTE_SRL	k0, k0, 6
	P_MTC0	k0, CP0_ENTRYLO0
	PTE_SRL	k1, k1, 6
	mfc0	k0, CP0_INDEX
	P_MTC0	k1, CP0_ENTRYLO1
	bltzl	k0, 1f
	tlbwr
1:
	nop
	eret
	END(except_vec0_r4k_mphwbug)
#endif

	/* TLB refill, EXL == 0, R4000 UP 250MHZ entrylo[01] hwbug version */
	LEAF(except_vec0_r4k_250MHZhwbug)
	.set	mips3
	mfc0	k0, CP0_BADVADDR
	srl	k0, k0, _PGDIR_SHIFT
	lw	k1, pgd_current			# get pgd pointer
	sll	k0, k0, 2			# log2(sizeof(pgd_t)
	addu	k1, k1, k0
	mfc0	k0, CP0_CONTEXT
	lw	k1, (k1)
#ifndef CONFIG_64BIT_PHYS_ADDR
	srl	k0, k0, 1
#endif
	and	k0, k0, PTEP_INDX_MSK
	addu	k1, k1, k0
	PTE_L	k0, 0(k1)
	PTE_L	k1, PTE_SIZE(k1)
	PTE_SRL	k0, k0, 6
	P_MTC0	zero, CP0_ENTRYLO0
	P_MTC0	k0, CP0_ENTRYLO0
	PTE_SRL	k1, k1, 6
	P_MTC0	zero, CP0_ENTRYLO1
	P_MTC0	k1, CP0_ENTRYLO1
	b	1f
	tlbwr
1:
	nop
	eret
	END(except_vec0_r4k_250MHZhwbug)

#ifdef CONFIG_SMP
	/* TLB refill, EXL == 0, R4000 MP 250MHZ entrylo[01]+badvaddr bug version */
	LEAF(except_vec0_r4k_MP250MHZhwbug)
	.set	mips3
	mfc0	k0, CP0_BADVADDR
	srl	k0, k0, _PGDIR_SHIFT
	lw	k1, pgd_current			# get pgd pointer
	sll	k0, k0, 2			# log2(sizeof(pgd_t)
	addu	k1, k1, k0
	mfc0	k0, CP0_CONTEXT
	lw	k1, (k1)
#ifndef CONFIG_64BIT_PHYS_ADDR
	srl	k0, k0, 1
#endif
	and	k0, k0, PTEP_INDX_MSK
	addu	k1, k1, k0
	PTE_L	k0, 0(k1)
	PTE_L	k1, PTE_SIZE(k1)
	nop				/* XXX */
	tlbp
	PTE_SRL	k0, k0, 6
	P_MTC0  zero, CP0_ENTRYLO0
	P_MTC0  k0, CP0_ENTRYLO0
	mfc0    k0, CP0_INDEX
	PTE_SRL	k1, k1, 6
	P_MTC0	zero, CP0_ENTRYLO1
	P_MTC0	k1, CP0_ENTRYLO1
	bltzl	k0, 1f
	tlbwr
1:
	nop
	eret
	END(except_vec0_r4k_MP250MHZhwbug)
#endif



	__FINIT

/*
 * ABUSE of CPP macros 101.
 *
 * After this macro runs, the pte faulted on is
 * in register PTE, a ptr into the table in which
 * the pte belongs is in PTR.
 */

#ifdef CONFIG_SMP
#define GET_PGD(scratch, ptr)        \
	mfc0    ptr, CP0_CONTEXT;    \
	la      scratch, pgd_current;\
	srl     ptr, 23;             \
	sll     ptr, 2;              \
	addu    ptr, scratch, ptr;   \
	lw      ptr, (ptr);
#else
#define GET_PGD(scratch, ptr)    \
	lw	ptr, pgd_current;
#endif

#define LOAD_PTE(pte, ptr) \
	GET_PGD(pte, ptr)          \
	mfc0	pte, CP0_BADVADDR; \
	srl	pte, pte, _PGDIR_SHIFT; \
	sll	pte, pte, 2; \
	addu	ptr, ptr, pte; \
	mfc0	pte, CP0_BADVADDR; \
	lw	ptr, (ptr); \
	srl	pte, pte, PTE_INDX_SHIFT; \
	and	pte, pte, PTE_INDX_MSK; \
	addu	ptr, ptr, pte; \
	PTE_L	pte, (ptr);

	/* This places the even/odd pte pair in the page
	 * table at PTR into ENTRYLO0 and ENTRYLO1 using
	 * TMP as a scratch register.
	 */
#ifndef CONFIG_BMIPS3300_WAR
#define PTE_RELOAD(ptr, tmp) \
	ori	ptr, ptr, PTE_SIZE; \
	xori	ptr, ptr, PTE_SIZE; \
	PTE_L	tmp, PTE_SIZE(ptr); \
	PTE_L	ptr, 0(ptr); \
	PTE_SRL	tmp, tmp, 6; \
	P_MTC0	tmp, CP0_ENTRYLO1; \
	PTE_SRL	ptr, ptr, 6; \
	P_MTC0	ptr, CP0_ENTRYLO0;
#else
#define PTE_RELOAD(ptr, tmp) \
	ori	ptr, ptr, PTE_SIZE; \
	xori	ptr, ptr, PTE_SIZE; \
	PTE_L	tmp, PTE_SIZE(ptr); \
	PTE_L	ptr, 0(ptr); \
	bltz	tmp, fixup_reload_odd; \
	PTE_SRL	tmp, tmp, 6; \
	P_MTC0	tmp, CP0_ENTRYLO1; \
	bltz	ptr, fixup_reload_even; \
	PTE_SRL	ptr, ptr, 6; \
	P_MTC0	ptr, CP0_ENTRYLO0;
#endif

#define DO_FAULT(write) \
	.set	noat; \
	SAVE_ALL; \
	mfc0	a2, CP0_BADVADDR; \
	KMODE; \
	.set	at; \
	move	a0, sp; \
	jal	do_page_fault; \
	 li	a1, write; \
	j	ret_from_exception; \
	 nop; \
	.set	noat;

	/* Check is PTE is present, if not then jump to LABEL.
	 * PTR points to the page table where this PTE is located,
	 * when the macro is done executing PTE will be restored
	 * with it's original value.
	 */
#define PTE_PRESENT(pte, ptr, label) \
	andi	pte, pte, (_PAGE_PRESENT | _PAGE_READ); \
	xori	pte, pte, (_PAGE_PRESENT | _PAGE_READ); \
	bnez	pte, label; \
	 PTE_L	pte, (ptr);

	/* Make PTE valid, store result in PTR. */
#define PTE_MAKEVALID(pte, ptr) \
	ori	pte, pte, (_PAGE_VALID | _PAGE_ACCESSED); \
	PTE_S	pte, (ptr);

	/* Check if PTE can be written to, if not branch to LABEL.
	 * Regardless restore PTE with value from PTR when done.
	 */
#define PTE_WRITABLE(pte, ptr, label) \
	andi	pte, pte, (_PAGE_PRESENT | _PAGE_WRITE); \
	xori	pte, pte, (_PAGE_PRESENT | _PAGE_WRITE); \
	bnez	pte, label; \
	 PTE_L	pte, (ptr);

	/* Make PTE writable, update software status bits as well,
	 * then store at PTR.
	 */
#define PTE_MAKEWRITE(pte, ptr) \
	ori	pte, pte, (_PAGE_ACCESSED | _PAGE_MODIFIED | \
			   _PAGE_VALID | _PAGE_DIRTY); \
	PTE_S	pte, (ptr);

	.set	noreorder

/*
 * From the IDT errata for the QED RM5230 (Nevada), processor revision 1.0:
 * 2. A timing hazard exists for the TLBP instruction.
 *
 *      stalling_instruction
 *      TLBP
 *
 * The JTLB is being read for the TLBP throughout the stall generated by the
 * previous instruction. This is not really correct as the stalling instruction
 * can modify the address used to access the JTLB.  The failure symptom is that
 * the TLBP instruction will use an address created for the stalling instruction
 * and not the address held in C0_ENHI and thus report the wrong results.
 *
 * The software work-around is to not allow the instruction preceding the TLBP
 * to stall - make it an NOP or some other instruction guaranteed not to stall.
 *
 * Errata 2 will not be fixed.  This errata is also on the R5000.
 *
 * As if we MIPS hackers wouldn't know how to nop pipelines happy ...
 */
#define R5K_HAZARD nop

	/*
	 * Note for many R4k variants tlb probes cannot be executed out
	 * of the instruction cache else you get bogus results.
	 */
	.align	5
	NESTED(handle_tlbl, PT_SIZE, sp)
	.set	noat
invalid_tlbl:
#ifdef TLB_OPTIMIZE
	/* Test present bit in entry. */
	LOAD_PTE(k0, k1)
	R5K_HAZARD
	tlbp
	PTE_PRESENT(k0, k1, nopage_tlbl)
	PTE_MAKEVALID(k0, k1)
	PTE_RELOAD(k1, k0)
	nop
	b	1f
	 tlbwi
1:
	nop
	.set	mips3
	eret
	.set	mips0
#endif

nopage_tlbl:
	DO_FAULT(0)
	END(handle_tlbl)

	.align	5
	NESTED(handle_tlbs, PT_SIZE, sp)
	.set	noat
#ifdef TLB_OPTIMIZE
	.set	mips3
        li      k0,0
	LOAD_PTE(k0, k1)
	R5K_HAZARD
	tlbp				# find faulting entry
	PTE_WRITABLE(k0, k1, nopage_tlbs)
	PTE_MAKEWRITE(k0, k1)
	PTE_RELOAD(k1, k0)
	nop
	b	1f
	 tlbwi
1:
	nop
	.set	mips3
	eret
	.set	mips0
#endif

nopage_tlbs:
	DO_FAULT(1)
	END(handle_tlbs)

	.align	5
	NESTED(handle_mod, PT_SIZE, sp)
	.set	noat
#ifdef TLB_OPTIMIZE
	.set	mips3
	LOAD_PTE(k0, k1)
	R5K_HAZARD
	tlbp					# find faulting entry
	andi	k0, k0, _PAGE_WRITE
	beqz	k0, nowrite_mod
	 PTE_L	k0, (k1)

	/* Present and writable bits set, set accessed and dirty bits. */
	PTE_MAKEWRITE(k0, k1)

	/* Now reload the entry into the tlb. */
	PTE_RELOAD(k1, k0)
	nop
	b	1f
	 tlbwi
1:
	nop
	.set	mips3
	eret
	.set	mips0
#endif

nowrite_mod:
	DO_FAULT(1)
	END(handle_mod)

#ifdef CONFIG_DEBUG_MEMPOOL
	.align 4
	.set	noreorder
	.set	noat
/* This is the optimized version of debug mempool, it can handle all
 * cases except :
 * a) Load/store in a branch delay
 * b) LL/SC or Load Linked/Store Conditional
 */

#if defined(CONFIG_CPU_MIPS32) || defined(CONFIG_CPU_MIPS64)
#define CP0_TEMP_REG CP0_DESAVE
#define CP0_TEMP2_REG CP0_ERROREPC
#else
#ifdef CONFIG_CPU_R5432
#define CP0_TEMP_REG CP0_LLADDR
#define CP0_TEMP2_REG CP0_ERROREPC
#else
#error "MEMPOOL_DEBUG is only supported for MIPS32 and R5432"
#endif
#endif
	
#define DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(reg) \
	jr	k0		; \
	mtc0	$reg, CP0_TEMP2_REG;

debug_mempool_store_register_table:
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(0)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(1)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(2)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(3)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(4)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(5)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(6)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(7)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(8)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(9)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(10)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(11)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(12)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(13)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(14)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(15)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(16)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(17)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(18)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(19)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(20)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(21)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(22)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(23)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(24)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(25)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(26)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(27)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(28)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(29)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(30)
	DEBUG_MEMPOOL_STORE_REGISTER_TABLE_ENTRY(31)


#define DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER \
	mfc0	k1, CP0_BADVADDR ; \
	lui	k0, 0x1000	 ; \
	addu	k1, k0		 ; \
	mfc0	k0, CP0_TEMP2_REG	 ; 
	
debug_mempool_store_opcode_table:
	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	return_debug_mempool
	sb	k0, 0(k1) //101000
	nop
	nop
	
	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	return_debug_mempool
	sh	k0, 0(k1) //101001
	nop
	nop
	
	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	return_debug_mempool
	swl	k0, 0(k1) //101010
	nop
	nop

	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	return_debug_mempool
	sw	k0, 0(k1) //101011
	nop
	nop

	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	goback_to_default_handler
	nop //101100
	nop
	nop

	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	goback_to_default_handler
	nop //101101
	nop
	nop

	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	return_debug_mempool
	swr	k0, 0(k1) //101110
	nop
	nop

	DEBUG_MEMPOOL_STORE_OPCODE_ENTRY_HEADER
	j	goback_to_default_handler
	nop	//101111
	nop
	nop
	
	
#define DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(reg) \
	j	return_debug_mempool ; \
	move	$reg, k1	     ;
	
debug_mempool_load_return_table:
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(0)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(1)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(2)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(3)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(4)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(5)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(6)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(7)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(8)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(9)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(10)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(11)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(12)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(13)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(14)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(15)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(16)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(17)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(18)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(19)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(20)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(21)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(22)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(23)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(24)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(25)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(26)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(27)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(28)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(29)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(30)
	DEBUG_MEMPOOL_LOAD_RETURN_TABLE_ENTRY(31)

#define DEBUG_MEMPOOL_LOAD_OPCODE_ENTRY_HEADER \
	mfc0	k1, CP0_BADVADDR ; \
	lui	k0, 0x1000	 ; \
	addu	k1, k0		 ; \
	mfc0	k0, CP0_TEMP_REG	 ; 
	
debug_mempool_load_opcode_table:
	DEBUG_MEMPOOL_LOAD_OPCODE_ENTRY_HEADER
	jr	k0
	lb	k1, 0(k1) //100000
	nop
	nop
	
	DEBUG_MEMPOOL_LOAD_OPCODE_ENTRY_HEADER
	jr	k0
	lh	k1, 0(k1) //100001
	nop
	nop

	j	debug_mempool_start_lwl //100010
	mfc0	k0, CP0_TEMP_REG
	nop
	nop
	nop
	nop
	nop
	nop

	DEBUG_MEMPOOL_LOAD_OPCODE_ENTRY_HEADER
	jr	k0
	lw	k1, 0(k1) //100011
	nop
	nop

	DEBUG_MEMPOOL_LOAD_OPCODE_ENTRY_HEADER
	jr	k0
	lbu	k1, 0(k1) //100100
	nop
	nop

	DEBUG_MEMPOOL_LOAD_OPCODE_ENTRY_HEADER
	jr	k0
	lhu	k1, 0(k1) //100101
	nop
	nop

	j	debug_mempool_start_lwr //100110
	mfc0	k0, CP0_TEMP_REG
	nop
	nop
	nop
	nop
	nop
	nop

	DEBUG_MEMPOOL_LOAD_OPCODE_ENTRY_HEADER
	j	goback_to_default_handler
	nop	//100111
	nop
	nop

debug_mempool_start_lwl:
	lui	k1, %hi(debug_mempool_load_return_table)
	addi	k1, %lo(debug_mempool_load_return_table)
	subu	k1, k0, k1
	lui	k0, %hi(debug_mempool_store_register_table)
	addi	k0, %lo(debug_mempool_store_register_table)
	addu	k1, k0
	lui	k0, %hi(debug_mempool_finish_lwl)
	jr	k1
	addi	k0, %lo(debug_mempool_finish_lwl)
debug_mempool_finish_lwl:
	mfc0	k1, CP0_BADVADDR
	lui	k0, 0x1000
	addu	k0, k1
	mfc0	k1, CP0_TEMP2_REG
	lwl	k1, 0 (k0)
	mfc0	k0, CP0_TEMP_REG
	jr	k0
	nop

debug_mempool_start_lwr:
	lui	k1, %hi(debug_mempool_load_return_table)
	addi	k1, %lo(debug_mempool_load_return_table)
	subu	k1, k0, k1
	lui	k0, %hi(debug_mempool_store_register_table)
	addi	k0, %lo(debug_mempool_store_register_table)
	addu	k1, k0
	lui	k0, %hi(debug_mempool_finish_lwr)
	jr	k1
	addi	k0, %lo(debug_mempool_finish_lwr)
debug_mempool_finish_lwr:
	mfc0	k1, CP0_BADVADDR
	lui	k0, 0x1000
	addu	k0, k1
	mfc0	k1, CP0_TEMP2_REG
	lwr	k1, 0 (k0)
	mfc0	k0, CP0_TEMP_REG
	jr	k0
	nop

	
	.align 4
debug_mempool_access:
	// This implicitly causes us to re-execute the USER code should a
	// fault occur, so this code *must* be re-entrant safe.
	mfc0	k1, CP0_CAUSE
	bltz	k1, goback_to_default_handler // We don't do branch delays
	 lw	k1, 0(k0)
	sll	k1, 1 // We handle everything that begins in binary 10
		      // This test steers us clear of the following insns:
		      // 1100XX LWCx
		      // 111000 SC
		      // 1110xx SWCx
		      // 110000 LL
	blez	k1, goback_to_default_handler
	 sll	k1, 1 // All faulting instructions of type store begin
		      // 101xxx, so we queue off this whether to go to
		      // to our store fastpath or our load fast path.
	bgez	k1, process_load_fastpath
	 mtc0	k1, CP0_TEMP_REG // Save it so we don't have to re-load it.

	// We're in a store pathway, need to check for deadbeef.
	
	// We are re-entrant at this point, if accessing the userval
	// causes a tlb exception, we'll return to the user instruction
	// and re-execute the codepath.  So, we don't bother with
	// fixup table BS.
	
	mfc0	k0, CP0_BADVADDR
	ori	k0, 0x3
	xori	k0, 0x3 // Word align the address
	lui	k1, 0x1000
	addu	k1, k0, k1 
	lw	k1, 0(k1)
	lui	k0, 0xdead
	ori	k0, 0xbeef
	beq	k1, k0, goback_to_default_handler
	 mfc0	k1, CP0_TEMP_REG // *EPC << 2
	/* For store I need:
	 * 1) In k0, address into store_opcode_table
	 * 2) In k1, address into store_register_table
	 */

	// Calculate address into store_opcode table
	srl	k1, 23 // Shift opcode into << 5 position, addresses 8 insns.
	andi	k1, 0x00e0 // And out 3 opcode bits we care about.
	lui	k0, %hi(debug_mempool_store_opcode_table)
	addi	k0, %lo(debug_mempool_store_opcode_table)
	addu	k0, k1

	mfc0	k1, CP0_TEMP_REG // *EPC << 2
	mtc0	k0, CP0_TEMP_REG

	// Calculate address into store_register_table
	srl	k1, 15  // Shift rt into rt << 3 position
	andi	k1, 0x00f8 // Mask out rt << 3
	lui	k0, %hi(debug_mempool_store_register_table)
	addi	k0, %lo(debug_mempool_store_register_table)
	addu	k1, k0


	// Jump off to store_register_table
	jr	k1
	mfc0	k0, CP0_TEMP_REG // Restore address into store_opcode_table

process_load_fastpath:
	srl	k1, 15  // rt starts at natural bit 18, we want this
			// upticked by 3, 2 for 4 bytes/insn,1 more for 2 insn.
			// We andi off only the addend we care about.
	andi	k1, 0x00f8
	lui	k0, %hi(debug_mempool_load_return_table)
	addi	k0, %lo(debug_mempool_load_return_table)
	addu	k0, k1
	
	mfc0	k1, CP0_TEMP_REG // EPC << 2
	mtc0	k0, CP0_TEMP_REG // Save our load return.

	srl	k1, 28 // Bit 31 will be 0 since we're load, the rest is
		       // zeroing on the right.
	sll	k1, 5  // 2 for 4 byte instructions, 3 for 8 instructions
	lui	k0, %hi(debug_mempool_load_opcode_table)
	addi	k0, %lo(debug_mempool_load_opcode_table)
	addu	k1, k0  //Jump to reg in k1	
	jr	k1
	nop

return_debug_mempool:
	mfc0	k0, CP0_EPC
	addiu	k0, k0, 4
	mtc0	k0, CP0_EPC
	nop			;  nop;
	.set mips3
	eret
	.set mips0
		 
goback_to_default_handler:
	// Jump to the general exception vector.
	j	0x180
	 mfc0	k0, CP0_BADVADDR
	.set	reorder
	.set	at
#endif
	
#ifdef CONFIG_BMIPS3300_WAR
	.align	4
	.set	noreorder
	.set	noat
fixup_multiple:
	SAVE_ALL
	mfc0	a1, CP0_BADVADDR
	KMODE
	.set	at
	jal	do_fixup_multiple
	 move	a0, sp
	j	ret_from_exception
	 nop
	.set	noat

	.align	4
fixup_reload_odd:
	/* Make sure previous PTE not conjoined to this one */
	mfc0	k0, CP0_BADVADDR
	lw	k1, pgd_current
	addiu	k0, -0x2000
	srl	k0, _PGDIR_SHIFT
	sll	k0, 2
	addu	k1, k0
	mfc0	k0, CP0_BADVADDR
	lw	k1, 0(k1)
	addiu	k0, -0x2000
	srl	k0, PTE_INDX_SHIFT
	and	k0, PTEP_INDX_MSK
	addu	k1, k0
	lw	k0, 4(k1)
	bgez	k0, fixup_reload_even
	 sll	k0,1
	bgez	k0, fixup_multiple
	 nop

fixup_reload_even:
	mfc0	k1, CP0_INDEX
	la	k0, KSEG0
	sll	k1, 13
	addu	k1, k0
	mfc0	k0, CP0_ENTRYHI
	mtc0	k1, CP0_ENTRYHI
	mtc0	zero, CP0_ENTRYLO1
	mtc0	zero, CP0_ENTRYLO0
	nop; nop
	tlbwi
	/* Rest of logic is same as for TLB miss */
	j	0
	 mtc0	k0, CP0_ENTRYHI

	.align	4
fixup_even_page:
	bltz	k0, 6f
	 sra	k0, 7
	mtc0	k0, CP0_ENTRYLO0
	andi	k0, 0x0002
	beqz	k0, 2f
	 andi	k0, k1, 0x0080

	/* even page valid */
	beqz	k0, 4f
	 andi	k0, k1, 0x0001
2:
	/* even invalid || (even valid && odd valid) */
	bltz	k1, fixup_odd_page
	 sll	k1, 1
	srl	k1, 7
	mtc0	k1, CP0_ENTRYLO1
3:
	b	1f
	 tlbwr
1:
	nop
	.set	mips3
	eret
	.set	mips0
4:
	/* even page valid, odd invalid */
	beqz	k0, 5f
	 mfc0	k0, CP0_BADVADDR

	/* odd page present, but not valid
	 *
	 * All we need to do here is or in some bits, but that
	 * needs to get written out to PTE, too, and we already
	 * threw out that pointer, so have to recompute it...
	 */
	lw	k1, pgd_current
	srl	k0, _PGDIR_SHIFT
	sll	k0, 2
	addu	k1, k0
	mfc0	k0, CP0_BADVADDR
	lw	k1, (k1)
	srl	k0, PTE_INDX_SHIFT
	and	k0, PTEP_INDX_MSK
	addu	k0, k1

	lw	k1, 4(k0)
	ori	k1, 0x0088
	b	2b
	 sw	k1, 4(k0)
5:
	b	fault_in_page
	 ori	k1, k0, 0x1000
6:
	sll	k0, 6
	srl	k0, 6
	b	2b
	 mtc0	k0, CP0_ENTRYLO0
7:
	sll	k1, 6
	srl	k1, 6
	b	3b
	 mtc0	k1, CP0_ENTRYLO1

	.align	4
fixup_odd_page:
	bltz	k1, 7b
	 sra	k1, 7
	mfc0	k0, CP0_ENTRYHI
	mtc0	k1, CP0_ENTRYLO1
	andi	k1, 0x0002

	/* don't bother if odd page invalid */
	beqz	k1, 3b
	 addiu	k0, 0x2000

	/* odd page valid */
	mtc0	k0, CP0_ENTRYHI
	nop; nop
	tlbp
	nop; nop
	mfc0	k1, CP0_INDEX
	bgez	k1, 4f
	 addiu	k0, -0x2000

	/* next page not in TLB */
	mtc0	k0, CP0_ENTRYHI
8:
	mfc0	k0, CP0_BADVADDR
	lw	k1, pgd_current
	addiu	k0, 0x2000
	srl	k0, _PGDIR_SHIFT
	sll	k0, 2
	addu	k1, k0
	mfc0	k0, CP0_BADVADDR
	lw	k1, 0(k1)
	addiu	k0, 0x2000
	srl	k0, PTE_INDX_SHIFT
	and	k0, PTEP_INDX_MSK
	addu	k1, k0

	/* fetch next PTE */
	lw	k0, 0(k1)
	andi	k0, 0x0001
	beqz	k0, 2f
	 lw	k0, 0(k1)

	/* next even page present */
	ori	k0, 0x0088
	sw	k0, 0(k1)
	bltz	k0, 6f
	 srl	k0, 6
7:
	lw	k1, 4(k1)
	bltz	k1, fixup_multiple
	 srl	k1, 6

	/* write out the first entry */
	tlbwr

	/* then do the next entry */
	mtc0	k0, CP0_ENTRYLO0
	mfc0	k0, CP0_ENTRYHI
	mtc0	k1, CP0_ENTRYLO1
	addiu	k0, 0x2000
	mtc0	k0, CP0_ENTRYHI
	nop; nop
	tlbwr
	nop; nop; nop
	.set	mips3
	eret
	.set	mips0

6:
	lw	k0, 4(k1)
	andi	k0, 0x0001
	beqz	k0, 3f
	 lw	k0, 4(k1)
	ori	k0, 0x0088
	sw	k0, 4(k1)
	lw	k0, 0(k1)
	sll	k0, 1
	b	7b
	 srl	k0, 7

4:
	/* next page in TLB, remove it */
	la	k0, KSEG0
	sll	k1, 13
	addu	k1, k0
	mfc0	k0, CP0_ENTRYHI
	mtc0	k1, CP0_ENTRYHI
	mtc0	zero, CP0_ENTRYLO1
	mtc0	zero, CP0_ENTRYLO0
	nop; nop
	tlbwi
	addiu	k0, -0x2000
	mtc0	k0, CP0_ENTRYHI
	/* need to recompute ENTRYLO0/ENTRYLO1 */
	mfc0	k0, CP0_BADVADDR
	lw	k1, pgd_current
	srl	k0, _PGDIR_SHIFT
	sll	k0, 2
	addu	k1, k0
	mfc0	k0, CP0_BADVADDR
	lw	k1, 0(k1)
	srl	k0, PTE_INDX_SHIFT
	and	k0, PTEP_INDX_MSK
	addu	k1, k0
	/* refetch PTE */
	lw	k0, 0(k1)
	lw	k1, 4(k1)
	sll	k0, 1
	sll	k1, 1
	srl	k0, 7
	srl	k1, 7
	mtc0	k0, CP0_ENTRYLO0
	b	8b
	 mtc0	k1, CP0_ENTRYLO1

3:
	/* next odd page not present */
	b	1f
	 li	k1, 0x2000

2:
	/* next even page not present */
	li	k1, 0x1000
1:
	mfc0	k0, CP0_BADVADDR
	ori	k0, 0x1000
	addu	k1, k0

fault_in_page:
	/* This is (non-SMP) SAVE_ALL, modified to preserve k1 */
	mfc0	k0, CP0_STATUS
	sll	k0, 3
	bltz	k0, 8f
	 move	k0, sp
	lui	k0, %hi(kernelsp)
	lw	k0, %lo(kernelsp)(k0)
8:
	sw	sp, (PT_R29 - PT_SIZE)(k0)
	subu	sp, k0, PT_SIZE
	sw	$3, PT_R3(sp)
	sw	$0, PT_R0(sp)
	mfc0	v1, CP0_STATUS
	sw	$2, PT_R2(sp)
	sw	v1, PT_STATUS(sp)
	sw	$4, PT_R4(sp)
	mfc0	v1, CP0_CAUSE
	sw	$5, PT_R5(sp)
	sw	v1, PT_CAUSE(sp)
	sw	$6, PT_R6(sp)
	mfc0	v1, CP0_EPC
	sw	$7, PT_R7(sp)
	sw	v1, PT_EPC(sp)
	sw	$25, PT_R25(sp)
	sw	$28, PT_R28(sp)
	sw	$31, PT_R31(sp)
#ifdef CONFIG_4KSTACK
	lui	v1, %hi(kernels28)
	lw	$28, %lo(kernels28)(v1)
#else
	ori	$28, sp, 0x1fff
	xori	$28, 0x1fff
#endif

	SAVE_AT
	SAVE_TEMP
	SAVE_STATIC

	move	a2, k1
	KMODE
	.set	at
	move	a0, sp
	jal	do_page_fault
	 move	a1, zero
	j	ret_from_exception
	 nop
	.set	noat
#endif

