//
//  IDE A/V command extension driver for 2.4 kernel
//
// Copyright (c) 2001, 2006 TiVo Inc.
//

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/hdreg.h>

#include <linux/ide.h>
#include <linux/ide-tivo.h>
#include <linux/tivo-private.h>

// #define DUMP_CONFIG
// #define WATCH_FLUSHES

#ifdef PERIODIC_MODE_LOG
static int swaps = 0;
#endif
#ifdef WATCH_FLUSHES
static int flushes = 0;
#endif

static int cautionlevel = 3;

extern void idetivo_set_statistics_hook(ide_startstop_t (*handler)(ide_drive_t *,
						        int, struct request *rq));
extern void idetivo_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *));

extern const int idetivo_version_supported;

#ifdef CONFIG_TCD_BSWAP
static void idetivo_bswap_data (void *buffer, int wcount)
{
    u16 *p = buffer;
   
    //    printk(KERN_INFO "Byteswapping %d words\n", wcount);

    while (wcount--) {
        *p++ = *p << 8 | *p >> 8;
        *p++ = *p << 8 | *p >> 8;
    }
}
#endif

static inline void idedisk_input_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	ide_input_data(drive, buffer, wcount);
#ifdef CONFIG_TCD_BSWAP
	idetivo_bswap_data(buffer, wcount);
#endif
}

static inline void idedisk_output_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
#ifdef CONFIG_TCD_BSWAP
  idetivo_bswap_data(buffer, wcount);
#endif
  ide_output_data(drive, buffer, wcount);
#ifdef CONFIG_TCD_BSWAP
  idetivo_bswap_data(buffer, wcount);
#endif
}

/*
 * getconfig_intr() is invoked on completion of a getconfig cmd.
 */
static ide_startstop_t getconfig_intr (ide_drive_t *drive)
{
   struct idetivo_drive_info *tivodisk = &drive->tivo;
   byte stat;
   stat = GET_STAT();

#ifdef DUMP_CONFIG
   printk(KERN_INFO "In getconfig_intr\n");
#endif
   if (!OK_STAT(stat,DATA_READY,BAD_STAT)) {
		ide_startstop_t	err = ide_error(drive, "getconfig_intr", stat);
      	return err;
   }
#ifdef DUMP_CONFIG
   printk("Getting config data from drive\n");
#endif
   idedisk_input_data(drive, tivodisk->config, SECTOR_WORDS);
#ifdef DUMP_CONFIG
   printk("Read %d words to 0x%p\n", SECTOR_WORDS, tivodisk->config);
#endif
   tivodisk->state = STATE_KNOWN;
   return ide_stopped;
}

/*
 * setconfig_intr() is invoked on completion of a setconfig cmd.
 */
static ide_startstop_t setconfig_intr (ide_drive_t *drive)
{
    byte stat;
    stat = GET_STAT();

#ifdef DUMP_CONFIG
    printk(KERN_INFO "In setconfig_intr\n");
#endif
    if (!OK_STAT(stat,READY_STAT,BAD_STAT)) {
        ide_startstop_t  err = ide_error(drive, "setconfig_intr", stat);
       return err;
    }
#ifdef DUMP_CONFIG
    printk(KERN_INFO "Sent config info to %s\n", drive->name);
#endif
    return ide_stopped;
}

/*
 * flush_intr() is invoked on completion of a flush cmd.
 */
static ide_startstop_t flush_intr (ide_drive_t *drive)
{
    byte stat;

    stat = GET_STAT();

#ifdef DUMP_CONFIG
    printk(KERN_INFO "In flush_intr\n");
#endif
    if (!OK_STAT(stat,READY_STAT,BAD_STAT)) {
      struct idetivo_drive_info *tivodisk = &drive->tivo;
      ide_startstop_t startstop;

      startstop = ide_error(drive, "flush_intr", stat);
      if (tivodisk && tivodisk->family == FAMILY_UNKNOWN) {
	printk(KERN_ERR "Drive %s doesn't implement FLUSH command??\n",
	       drive->name);
	tivodisk->state = STATE_NOMEDIA;
      }
      return startstop;
    }
#ifdef DUMP_CONFIG
    printk(KERN_INFO "Flushed buffer on %s\n", drive->name);
#endif
    return ide_stopped;
}

/*
 * misc_intr() is invoked on completion of miscellaneous cmds.
 */
static ide_startstop_t misc_intr (ide_drive_t *drive)
{
    byte stat;

    stat = GET_STAT();

#ifdef DUMP_CONFIG
    printk(KERN_INFO "In misc_intr\n");
#endif
    if (!OK_STAT(stat,READY_STAT,BAD_STAT)) {
       ide_startstop_t err =  ide_error(drive, "misc_intr", stat);
       return err;
    }
    return ide_stopped;
}

static void initiate_flush (ide_drive_t *drive)
{
   unsigned long flags;

   save_flags(flags);
   sti();
   sata_pm_lock_port(drive);
     {
	byte stat;
	stat = GET_STAT();
	if (stat & BUSY_STAT) {
	   printk(KERN_INFO "ideturbo: initiate_flush with drive in use, stat 0x%X\n", stat);
	}
     }
   OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
#ifdef CONFIG_TIVO_SATAPM
	if ((drive->ctl) & 0x1)
		sata_pm_ide_intr_pending = 0;
	else
		sata_pm_ide_intr_pending = 1;
#endif // CONFIG_TIVO_SATAPM
   TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(drive->select.b.unit ? 0xB0 : 0xA0); /* drive */
   ide_set_handler (drive, &flush_intr, WAIT_CMD, NULL);
#ifdef DUMP_CONFIG
   printk("ideturbo: flushing buffer on %s\n", drive->name);
#endif
   TIVO_OUT_BYTE_IDE_COMMAND_REG_VALUE(WIN_FLUSH_CACHE);
   restore_flags(flags);
#ifdef WATCH_FLUSHES
   flushes++;
   if ((flushes % 100) == 0) {
      printk("ideturbo: %d flushes\n", flushes);
   }
#endif   
}

static inline ide_startstop_t ide_genericstats(ide_drive_t *drive,
					       int reset_statistics,
					       struct request *rq)
{
   struct idetivo_drive_info *tivodisk = &drive->tivo;

   sata_pm_lock_port(drive);
   switch (tivodisk->state) {
   case STATE_UNKNOWN:
     tivodisk->back_to_back_limit = 1;
     tivodisk->state = STATE_MEDIA;
     /* drop through */
   case STATE_MEDIA:
     if (rq->cmd == READ) {
       tivodisk->back_to_back_writes = 0;
     } else {
       tivodisk->back_to_back_writes ++;
       if (tivodisk->back_to_back_writes > tivodisk->back_to_back_limit) {
	 tivodisk->back_to_back_writes = 0;
	 initiate_flush(drive);
	 return ide_started;
       }
     }
   	 sata_pm_unlock_port(drive);
     return ide_stopped;
   case STATE_NOMEDIA:
   default:
   	 sata_pm_unlock_port(drive);
     return ide_stopped;
   }
}

static inline ide_startstop_t ide_tivoavstats(ide_drive_t *drive,
					      int reset_statistics,
					      struct request *rq)
{
   struct idetivo_drive_info *tivodisk = &drive->tivo;
   unsigned long flags;
   ide_startstop_t startstop;

   sata_pm_lock_port(drive);
   switch (tivodisk->state) {
   case STATE_UNKNOWN:
     tivodisk->back_to_back_limit = 1;
     tivodisk->state = STATE_NOMEDIA;
     printk(KERN_INFO "%s: Enabling write cache\n", drive->name);
     save_flags(flags);
     sti();
     {
       byte stat;
       stat = GET_STAT();
       if (stat & BUSY_STAT) {
	 printk(KERN_INFO "ideturbo: entered with drive in use, stat 0x%X\n",
		stat);
       }
     }
     OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
#ifdef CONFIG_TIVO_SATAPM
	if ((drive->ctl) & 0x1)
		sata_pm_ide_intr_pending = 0;
	else
		sata_pm_ide_intr_pending = 1;
#endif // CONFIG_TIVO_SATAPM
     TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(drive->select.b.unit ? 0x10 : 0x00); /* drive */
     ide_set_handler (drive, &misc_intr, WAIT_CMD, NULL);
     OUT_BYTE(0x02, IDE_FEATURE_REG);
     TIVO_OUT_BYTE_IDE_COMMAND_REG_VALUE(0xEF);
     restore_flags(flags);
     return ide_started; /* wait for interrupt before proceeding */
    case STATE_NOMEDIA:
     memset(tivodisk->config, 0, sizeof tivodisk->config);
     /*
       Change the following as necessary.  Note that the WD drives
       organize data in terms of shortwords, not bytes (I missed this
       initially!).  Assume big-endian byte ordering.
     */
     tivodisk->config[1] =  1;
     tivodisk->config[9] =  255;
     tivodisk->config[11] =  255;
     tivodisk->config[13] =  2;
     /*
       See above
     */
     printk(KERN_INFO "%s: Setting new A/V configuration:\n", drive->name);
     printk(KERN_INFO " A/V features enabled:      %d\n",
	    tivodisk->config[1]&1);
     printk(KERN_INFO " A/V read recovery time:    %d msec\n",
	    tivodisk->config[9]);
     printk(KERN_INFO " A/V write recovery time:   %d msec\n",
	    tivodisk->config[11]);
     printk(KERN_INFO " A/V max streams:           %d %s\n",
	    tivodisk->config[13] & 0x7,
	    (tivodisk->config[13] & 0x7) ? "" : "unlimited");
     tivodisk->state = STATE_MEDIA;
     save_flags(flags);
     sti();
     {
       byte stat;
       stat = GET_STAT();
       if (stat & BUSY_STAT) {
	 printk(KERN_INFO "ideturbo: entered with drive in use, stat 0x%X\n",
		stat);
       }
     }
     OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
#ifdef CONFIG_TIVO_SATAPM
	if ((drive->ctl) & 0x1)
		sata_pm_ide_intr_pending = 0;
	else
		sata_pm_ide_intr_pending = 1;
#endif // CONFIG_TIVO_SATAPM
     TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(drive->select.b.unit ? 0x10 : 0x00); /* drive */
     ide_set_handler (drive, &setconfig_intr, WAIT_CMD, NULL);
     TIVO_OUT_BYTE_IDE_COMMAND_REG_VALUE(0x80);
     if (ide_wait_stat(&startstop, drive, DATA_READY, drive->bad_wstat, WAIT_DRQ)) {
       printk(KERN_ERR "%s: no DRQ after issuing set-configuration command\n",
	      drive->name);
	   if (startstop == ide_stopped)
   			sata_pm_unlock_port(drive);
       return startstop;
     }
     idedisk_output_data(drive, tivodisk->config, SECTOR_WORDS);
     restore_flags(flags);
     return ide_started; /* wait for interrupt before proceeding */
   case STATE_MEDIA:
   default:
     if (rq->cmd == READ) {
       tivodisk->back_to_back_writes = 0;
     } else {
       tivodisk->back_to_back_writes ++;
       if (tivodisk->back_to_back_writes > tivodisk->back_to_back_limit) {
	 tivodisk->back_to_back_writes = 0;
	 initiate_flush(drive);
	 return ide_started;
       }
     }
   	 sata_pm_unlock_port(drive);
     return ide_stopped;
   }
}

static inline ide_startstop_t ide_tivoquantumstats(ide_drive_t *drive,
						   int reset_statistics,
						   struct request *rq)
{
   struct idetivo_drive_info *tivodisk = &drive->tivo;
   enum media_state desired_state;
#ifdef DUMP_CONFIG
   char strbuf[80];
   int i, j;
   char hexen[] = "0123456789ABCDEF";
#endif
   unsigned long flags;
   ide_startstop_t startstop;

   sata_pm_lock_port(drive);
   if (tivodisk->state == STATE_UNKNOWN) {
      tivodisk->back_to_back_limit = 1;
      printk(KERN_INFO "Need to get drive statistics\n");
      save_flags(flags);
      sti();
      OUT_BYTE(drive->ctl,IDE_CONTROL_REG); /* turn off interrupt */
#ifdef CONFIG_TIVO_SATAPM
		if ((drive->ctl) & 0x1)
			sata_pm_ide_intr_pending = 0;
		else
			sata_pm_ide_intr_pending = 1;
#endif // CONFIG_TIVO_SATAPM
      TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(drive->select.b.unit ? 0xB0 : 0xA0); /* drive */
      OUT_BYTE(0x01, IDE_NSECTOR_REG); /* get config, no save */
      OUT_BYTE(0xFF, IDE_SECTOR_REG); /* password */
      OUT_BYTE(0xFF, IDE_LCYL_REG); /* password */
      OUT_BYTE(0x3F, IDE_HCYL_REG); /* password */
      ide_set_handler (drive, &getconfig_intr, WAIT_CMD, NULL);
      TIVO_OUT_BYTE_IDE_COMMAND_REG_VALUE(0xF0);
      restore_flags(flags);
      return ide_started; /* go around again... */
   }
#if 0
   if (rq->tivo && (rq->tivo->partition_flags & 1)) {
      desired_state = STATE_MEDIA;
   } else {
      desired_state = STATE_NOMEDIA;
   }
#else
   /* Temporary kludge to get around possible drive microcode problem with
      frequent mode changes.  Force application mode. */
   desired_state = STATE_NOMEDIA;
#endif
   if (rq->cmd == READ) {
     tivodisk->back_to_back_writes = 0;
   } else {
     tivodisk->back_to_back_writes ++;
     if (tivodisk->back_to_back_writes > tivodisk->back_to_back_limit) {
       tivodisk->back_to_back_writes = 0;
       initiate_flush(drive);
       return ide_started;
     }
   }
   if (tivodisk->state == desired_state) {
   	  sata_pm_unlock_port(drive);
      return ide_stopped;
   }
   if (tivodisk->state != STATE_FLUSHED) {
      initiate_flush(drive);
      tivodisk->state = STATE_FLUSHED;
      return ide_started;
   }
#ifdef DUMP_CONFIG
   printk (KERN_INFO "Switching %s to %s mode\n", drive->name,
	   (desired_state == STATE_MEDIA) ? "media" : "application");
#endif
   tivodisk->config[32] = 0x3; /* prefetch enable, read cache enable */
   switch (desired_state) {
    case STATE_MEDIA:
      switch (cautionlevel) {
      case 0:
	/* turn on RC, turn off AWRE, ARRE, EEC, DCR */
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x10;
	/* turn retries down to zero */
	tivodisk->config[37] = 0;
	break;
      case 1:
	/* turn off RC, DCR, AWRE, ARRE, DCR, turn on EEC */
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x08;
	/* turn retries down to zero */
	tivodisk->config[37] = 0;
	break;
      case 2:
	/* turn off RC, DCR, AWRE, ARRE, DCR, turn on EEC */
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x08;
	/* Set for one retry sequence */
	tivodisk->config[37] = 1;
	break;
      default:
	/* turn off RC, DCR, AWRE, ARRE, DCR, turn on EEC */
	tivodisk->config[36] = (tivodisk->config[36] & 0x26) | 0x08;
	/* allow two retry cycles */
	tivodisk->config[37] = 2;
	break;
      }
      /* turn on write cache, turn off ruee */
      tivodisk->config[39] = (tivodisk->config[39] & 0xFD) | 0x4;
      break;
    case STATE_NOMEDIA:
      /* turn off RC and DCR, turn on AWRE, ARRE, EEC */
      tivodisk->config[36] = (tivodisk->config[36] & 0xEE) | 0xC8;
      /* allow two retry cycles */
      tivodisk->config[37] = 2;
      /* turn on write cache and ruee */
      tivodisk->config[39] = tivodisk->config[39] | 0x6;
      break;
    default:
      break;
   }
#ifdef DUMP_CONFIG
   for (i = 0, j = 0 ; i < 64 ; i++) {
      if ((i % 16) == 0) {
	 if (i > 0) {
	    strbuf[j]=0;
	    printk (KERN_INFO "%s\n", strbuf);
	 }
	 j=0;
      }
      strbuf[j++] = hexen[0xF & (tivodisk->config[i] >> 4)];
      strbuf[j++] = hexen[0xF & tivodisk->config[i]];
      strbuf[j++] = ' ';
   }
   strbuf[j]=0;
   printk (KERN_INFO "%s\n", strbuf);
#endif
#ifdef PERIODIC_MODE_LOG
   if ((++swaps % 1000) == 0) {
      printk(KERN_INFO "ideturbo: %d mode changes\n", swaps);
   }
#endif
   save_flags(flags);
   sti();
     {
	byte stat;
	stat = GET_STAT();
	if (stat & BUSY_STAT) {
	   printk(KERN_INFO "ideturbo: entered with drive in use, stat 0x%X\n", stat);
	}
     }
   OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
#ifdef CONFIG_TIVO_SATAPM
	if ((drive->ctl) & 0x1)
		sata_pm_ide_intr_pending = 0;
	else
		sata_pm_ide_intr_pending = 1;
#endif // CONFIG_TIVO_SATAPM
   TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(drive->select.b.unit ? 0xB0 : 0xA0); /* drive */
   OUT_BYTE(0xFE, IDE_NSECTOR_REG); /* set config, no save */
   OUT_BYTE(0xFF, IDE_SECTOR_REG); /* password */
   OUT_BYTE(0xFF, IDE_LCYL_REG); /* password */
   OUT_BYTE(0x3F, IDE_HCYL_REG); /* password */
   ide_set_handler (drive, &setconfig_intr, WAIT_CMD, NULL);
   TIVO_OUT_BYTE_IDE_COMMAND_REG_VALUE(0xF0);
   if (ide_wait_stat(&startstop, drive, DATA_READY, drive->bad_wstat, WAIT_DRQ)) {
      printk(KERN_ERR "%s: no DRQ after issuing set-configuration command\n", drive->name);
   	  sata_pm_startstop_unlock_port(startstop,drive);
      return startstop;
   }
#ifdef NOTDEF
   if (!drive->unmask)
     cli();
#endif
   idedisk_output_data(drive, tivodisk->config, SECTOR_WORDS);
   tivodisk->state = desired_state;
#ifdef DUMP_CONFIG
   printk (KERN_INFO "Sent data to switch to %s mode %s\n", drive->name,
	   (desired_state == STATE_MEDIA) ? "media" : "application");
#endif
   restore_flags(flags);
   return ide_started; /* wait for interrupt before proceeding */
}


ide_startstop_t ide_tivostats(ide_drive_t *drive,
			      int reset_statistics,
			      struct request *rq)
{
   struct idetivo_drive_info *tivodisk = &drive->tivo;

   sata_pm_lock_port(drive);
   if (IDE_TIVO_INFO->version != IDETIVO_DRIVER_VERSION) {
     printk(KERN_ERR "ideturbo version mismatch, disabling\n");
     idetivo_set_statistics_hook(NULL);
     idetivo_set_cmd_hook(NULL);
     sata_pm_unlock_port(drive);
     return ide_stopped;
   }
   // If a TiVo IO is IDETIVO_INPROGRESS, that means this is a continuation
   // of a request already in progress (because the number of sectors was
   // > the max number of sectors per transfer), and we were already called
   // before the request was started
   if (rq->tivo && rq->tivo->status == IDETIVO_INPROGRESS) {
     sata_pm_unlock_port(drive);
     return ide_stopped;
   }
   if (!tivodisk) {
     sata_pm_unlock_port(drive);
     return ide_stopped;
   }
   if (tivodisk->family == FAMILY_UNKNOWN) {
     if (drive->id->model[0] == 'Q' &&
	 drive->id->model[1] == 'U' &&
	 drive->id->model[2] == 'A') {
       printk(KERN_INFO "%s: Quantum management\n", drive->name);
       tivodisk->family = FAMILY_QUANTUM;
     } else if (drive->id->model[0] == 'W' &&
		drive->id->model[1] == 'D' &&
		(drive->id->command_set_2 != 0x0000 &&
		 drive->id->command_set_2 != 0xffff &&
		 drive->id->command_set_2 & 0x0080)) {
       printk(KERN_INFO "%s: StreamWeaver management\n", drive->name);
       tivodisk->family = FAMILY_STREAMWEAVER1;
     } else {
       printk(KERN_INFO "%s: Generic ATA management\n", drive->name);
       tivodisk->family = FAMILY_UNSUPPORTED;
     }
   }

   if (reset_statistics) {
      tivodisk->state = STATE_UNKNOWN;
      sata_pm_unlock_port(drive);
      return ide_stopped;
   }

   switch (tivodisk->family) {
   default:
     printk(KERN_WARNING "%s: invalid drive family set\n", drive->name);
     // fall through
   case FAMILY_UNSUPPORTED:
     return sata_pm_startstop_unlock_port(ide_genericstats(drive, reset_statistics, rq),drive);
   case FAMILY_STREAMWEAVER1:
     return sata_pm_startstop_unlock_port(ide_tivoavstats(drive, reset_statistics, rq),drive);
   case FAMILY_QUANTUM:
     return sata_pm_startstop_unlock_port(ide_tivoquantumstats(drive, reset_statistics, rq),drive);
   }
}

void ide_tivocmdhook(ide_drive_t *drive, unsigned char *cmdcode)
{
#ifdef CONFIG_TCD   
  struct idetivo_request *req; 
  static int mutate = 0, total = 0;

  if (IDE_TIVO_INFO->version != IDETIVO_DRIVER_VERSION) {
    printk(KERN_ERR "ideturbo version mismatch, disabling\n");
    idetivo_set_statistics_hook(NULL);
    idetivo_set_cmd_hook(NULL);
    return;
  }
  if (IDE_TIVO_DRIVE_INFO->family != FAMILY_STREAMWEAVER1) {
    return;  /* No A/V command set, don't frob command code */
  }
  total++;
#ifdef NOTDEF
  if ((total % 1000) == 0) {
    printk (KERN_INFO "A/V mode: %d total, %d mutations\n", total, mutate);
  }
#endif
  req = HWGROUP(drive)->rq->tivo;
  if (!req) {
    return;    /* Not a readsectors/writesectors, don't frob command code */
  }
  if (!(req->partition_flags & 1)) {
    return;  /* Not on a media partition, don't frob command code */
  }
  /* 
     TiVo I/O request, on a media partition, on a drive which supports
     the A/V stream commands.  Transform normal read/write command code
     into equivalent media command.  Set the IDE feature register to
     indicate a non-urgent, non-read-continuous request.  Use a stream
     ID of 0 for reads, and a stream ID of 1 for writes.
  */
  sata_pm_lock_port(drive);
  mutate ++;
  switch (*cmdcode) {
  case WIN_READ:
    *cmdcode = WIN_READAV;
    OUT_BYTE(0x00, IDE_FEATURE_REG);
    break;
  case WIN_READDMA:
    *cmdcode = WIN_READAVDMA;
    OUT_BYTE(0x00, IDE_FEATURE_REG);
    break;
  case WIN_WRITE:
    *cmdcode = WIN_WRITEAV;
    OUT_BYTE(0x01, IDE_FEATURE_REG);
    break;
  case WIN_WRITEDMA:
    *cmdcode = WIN_WRITEAVDMA;
    OUT_BYTE(0x01, IDE_FEATURE_REG);
    break;
  default:
    break;
  }
 sata_pm_unlock_port(drive);
#else
  (void) drive;
  (void) cmdcode;
#endif     
}


static int ide_turbo_init(void)
{
   if (idetivo_version_supported != IDETIVO_DRIVER_VERSION) {
    printk(KERN_ERR "ideturbo version mismatch, not enabling\n");
    return 0;
   }
   idetivo_set_statistics_hook(ide_tivostats);
   idetivo_set_cmd_hook(ide_tivocmdhook);
#ifdef CONFIG_TCD_BSWAP
   printk(KERN_INFO "TiVo disk statistics module loaded (byteswap enabled)\n");
#else
   printk(KERN_INFO "TiVo disk statistics module loaded\n");
#endif
   printk(KERN_INFO "Disk cautionlevel is %d\n", cautionlevel);
   return 0;
}

static void ide_turbo_cleanup(void)
{
   idetivo_set_statistics_hook(NULL);
   idetivo_set_cmd_hook(NULL);
   printk(KERN_INFO "TiVo disk statistics module unloaded\n");
}
module_init(ide_turbo_init);

EXPORT_NO_SYMBOLS;
