/*
 * Tivo modified version of Broadcom drivers/usb/brcmusb.h
 *
 /7401* Copyright (C) 2001, 2004, 2007 Broadcom Corporation
 *                    Sasi Kumar <sasik@broadcom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * BRCM adaption of the usb-ohci stack to support the direct (non-PCI) 
 * USB Host Controllers.  
 *
 * Include this file in usb-ohci.c BEFORE including usb-ohci.h.  This is
 * done to fake out the check for the CONFIG_PCI macro and avoid having 
 * to change the usb-ohci.h file!
 * 
 * 08-01-2004   MM     Added support for BCM7038
 * 02-08-2002   SK     Created
 */

#ifdef CONFIG_HAS_BRCM_USB

#ifndef __BCMUSB_H__
#define __BCMUSB_H__

#if defined(CONFIG_MIPS_BCM7320) || defined(CONFIG_MIPS_BCM7110)
#error "BCM7320/BCM7110 not supported"
#endif

// Fusion
#ifdef CONFIG_TIVO_FUSION

#include <asm/bcm7401.h>

// USB controller definitions,  physical address are for PCI
#define HC_BASE_ADDR                0x10480400              // physical!
#define HC_END_ADDR                 (HC_BASE_ADDR+0x6f)     
#define HC_INT_VECTOR               BCM_IRQ_USB_OHCI_0_CPU_INTR

#define HC2_BASE_ADDR               0x10480500              // physical!
#define HC2_END_ADDR                (HC2_BASE_ADDR+0x6f)    
#define HC2_INT_VECTOR              BCM_IRQ_USB_OHCI_1_CPU_INTR

#define EHC_BASE_ADDR               0x10480300              // physical!
#define EHC_END_ADDR                (EHC_BASE_ADDR+0xa8)
#define EHC_INT_VECTOR              BCM_IRQ_USB_EHCI_CPU_INTR 

// Broadcom USB control register definition
#define BrcmUsbSetup                0xb0480228             
#ifdef CONFIG_MIPS_BRCM974XX
#define BRCM_USB_SETUP_REG_VAL      (IOC+WABO_LE+FNBO_REV+FNHW_LE+IPP) // power control is negative logic
#else
#define BRCM_USB_SETUP_REG_VAL      (IOC+WABO_LE+FNBO_REV+FNHW_LE)
#endif

// Broadcom PLL control reg definitions
#define BrcmPllControl              0xb04801e8             
#define BRCM_PLL_CONTROL_REG_VAL    0x8  //use external clock until internal is qualified


#define DRIVER_MODS                 "TiVo Fusion USB Host Controller"

// #########

// Eiger
#elif defined CONFIG_TIVO_EIGER

#include <asm/bcm7038.h>

#ifndef CONFIG_MIPS_BCM7038C0
        #error "USB only supported on 7038C0"
#endif /* CONFIG_MIPS_BCM7038A0 */

//#ifdef CONFIG_MIPS_BCM7038C0

#define USB_HOST_ADR_BASE           (BCM_REG_BASE+0x004c0400)
#define BCM_LINUX_USB_HOST_IRQ      BCM_IRQ_USB_CPU_INTR
#define HC_BASE_ADDR                (USB_HOST_ADR_BASE-0xA0000000)
#define HC_END_ADDR                 (HC_BASE_ADDR+0x6f)
#define HC_INT_VECTOR               BCM_IRQ_USB_OHCI_0_CPU_INTR
#define EHC_BASE_ADDR               (USB_HOST_ADR_BASE-0x100-0xA0000000)
#define EHC_END_ADDR                (EHC_BASE_ADDR+0xa8)
#define EHC_INT_VECTOR              BCM_IRQ_USB_EHCI_CPU_INTR 

//#else

//#define USB_HOST_ADR_BASE           (BCM_REG_BASE+0x004c0400)
//#define BCM_LINUX_USB_HOST_IRQ      BCM_IRQ_USB_CPU_INTR
//#define HC_BASE_ADDR                (USB_HOST_ADR_BASE-0xA0000000)
//#define HC_END_ADDR                 (HC_BASE_ADDR+0xfff)
//#define HC_INT_VECTOR               BCM_IRQ_USB_CPU_INTR 
//#define EHC_BASE_ADDR               (USB_HOST_ADR_BASE-0x100-0xA0000000)
//#define EHC_END_ADDR                (EHC_BASE_ADDR+0xa4)
//#define EHC_INT_VECTOR              BCM_IRQ_USB_CPU_INTR

//#endif

#define BrcmPllControl              (BCM_REG_BASE+0x004c01e8)
#define BrcmUsbSetup                (BCM_REG_BASE+0x004c0228)

#define BRCM_USB_SETUP_REG_VAL      (IOC+WABO_LE+FNBO_REV+FNHW_LE)
#define BRCM_PLL_CONTROL_REG_VAL    (0x8) //use external clock until internal is qualified

#define DRIVER_MODS                 "TiVo Eiger USB Host Controller"

// #########

// UMA2C  
#elif defined(CONFIG_TIVO_UMA2C)
#include <asm/bcm7315.h>
#define TM_TOP_ADR_BASE             (BCM7315_REG_BASE+0x8000)
#ifdef CONFIG_TGC
#define USB_HOST_ADR_BASE	    (BCM7315_REG_BASE+0x8200)
#else
#define USB_HOST_ADR_BASE           (BCM7315_REG_BASE+0x8100)
#endif
#define HC_BASE_ADDR                USB_HOST_ADR_BASE
#define HC_END_ADDR                 (HC_BASE_ADDR+0x6f)
#define HC_INT_VECTOR               BCM_LINUX_USB_HOST_IRQ
#define EHC_BASE_ADDR               (USB_HOST_ADR_BASE+0x70)
#define EHC_END_ADDR                (EHC_BASE_ADDR+0x8f)
#define EHC_INT_VECTOR              BCM_LINUX_USB_HOST_IRQ

#define BrcmPllControl              (USB_HOST_ADR_BASE+0xf4)
#define BrcmUsbSetup                (USB_HOST_ADR_BASE+0xfc)

#define BRCM_USB_SETUP_REG_VAL      (IOC+WABO_LE+FNBO_REV+FNHW_LE)
#ifdef CONFIG_TGC
#define BRCM_PLL_CONTROL_REG_VAL    (0x7a)
#else
#define BRCM_PLL_CONTROL_REG_VAL    (0xfd) //use external clock until internal
                                           //is qualified
#endif

#define DRIVER_MODS                 "TiVo 2-chip USB Host Controller"

//#if defined(CONFIG_MIPS_BCM7320)
// #define HC2_BASE_ADDR              USB_HOST2_ADR_BASE
// #define HC2_END_ADDR               (USB_HOST2_ADR_BASE+0xfff)
// #define HC2_INT_VECTOR             BCM_LINUX_USB_HOST2_IRQ
//#endif

#else

#error "Platform is not supported"

#endif

#ifndef CONFIG_PCI
  #define CONFIG_PCI
#endif

#include <linux/pci.h>


enum { BABO_LE=0x01, FNHW_LE=0x02, FNBO_REV=0x04, WABO_LE=0x08,
       IOC=0x10, IPP=0X20, CLV_BE=0x200000 };

//**********************************************************************************
//some backport fixups...
#ifndef __list_splice
 #define __list_splice          list_splice
#endif
#ifndef BUG_ON
 #define BUG_ON(condition) do { if (unlikely((condition)!=0)) BUG(); } while(0)
#endif


#ifdef EHCI_BRCM
//some forward decls...
static const struct hc_driver ehci_driver;
int usb_hcd_pci_probe (struct pci_dev *dev, const struct pci_device_id *id);
void usb_hcd_pci_remove (struct pci_dev *dev);

#ifdef CONFIG_TIVO_UMA2C // for BCM7317 only    
#ifdef CONFIG_TGC
#define WRITE_VENDOR_SPECIFIC_REGISTERS
#else
#define WRITE_VENDOR_SPECIFIC_REGISTERS                \
    writel( 0x007f007f, (volatile u32 *) 0xfffe81a4 ); \
    writel( 0x00000001, (volatile u32 *) 0xfffe81ac );
#endif // CONFIG_TGC
#endif

static const struct pci_device_id __devinitdata brcm_ehci_pci_ids [] = 
{ 
        {

    /* handle any USB EHCI controller */
    .class =      ((PCI_CLASS_SERIAL_USB << 8) | 0x20),
    .class_mask =     ~0,
        .driver_data =  (unsigned long) &ehci_driver,

    /* no matter who makes it */
    vendor:     PCI_ANY_ID,
    device:     PCI_ANY_ID,
    subvendor:  PCI_ANY_ID,
    subdevice:  PCI_ANY_ID,

    }, { /* end: all zeroes */ }
};

struct pci_dev *Pci_ehci_dev;
/* --------------------------------------------------------------------------
    Name: brcm_ehcd_init
 -------------------------------------------------------------------------- */

static int brcm_ehcd_init (void)
{
    int status = 0;

    // Init BRCM USB setup registers for board and HC specific issues
    writel( BRCM_USB_SETUP_REG_VAL, BrcmUsbSetup );
    writel( BRCM_PLL_CONTROL_REG_VAL, BrcmPllControl );

#ifdef CONFIG_TIVO_UMA2C  //for BCM7317 only
#ifdef CONFIG_TGC
    //set sync_det_length (USB_UTMICTRL) (from Broadcom)
    writel( 0xff060140, (volatile u32 *) 0xfffe82ec );
#else
    //straighten out frame length
    writel( 0x000c0020, (volatile u32 *) 0xfffe81f8 );

    //set sync_det_length (USB_UTMICTRL)
    writel( 0xff060100, (volatile u32 *) 0xfffe81ec );
#endif // CONFIG_TGC
#endif // CONFIG_TIVO_UMA2C
    
#ifdef CONFIG_TIVO_BCM7038
    /* Force the controllers through reset to fix the lockup on 7038B1 during reboot */
    *((volatile u32 *) 0xb04c0310) = 2;
    *((volatile u32 *) 0xb04c0408) = 1;
    *((volatile u32 *) 0xb04c0508) = 1;
#endif       

    Pci_ehci_dev = kmalloc( sizeof( struct pci_dev ), GFP_KERNEL );
    if( !Pci_ehci_dev )
        return -ENOMEM;
    strcpy( Pci_ehci_dev->name, "BRCM-EHCI" );
    strcpy( Pci_ehci_dev->slot_name, "EHCI-Direct" );
    pci_resource_start(Pci_ehci_dev, 0) = EHC_BASE_ADDR;
    pci_resource_end(Pci_ehci_dev, 0) = EHC_END_ADDR;
    Pci_ehci_dev->irq = EHC_INT_VECTOR;
    status = usb_hcd_pci_probe (Pci_ehci_dev, &brcm_ehci_pci_ids[0]);

    return( status );
}

/* --------------------------------------------------------------------------
    Name: brcm_ehcd_cleanup
 -------------------------------------------------------------------------- */

static void brcm_ehcd_cleanup (void)
{
    usb_hcd_pci_remove( Pci_ehci_dev );
    if( Pci_ehci_dev )
        kfree( Pci_ehci_dev );
}

#elif defined(OHCI_BRCM)

//forward decls
static struct pci_driver ohci_pci_driver;
static int __devinit ohci_pci_probe (struct pci_dev *dev, const struct pci_device_id *id);
static void __devexit ohci_pci_remove (struct pci_dev *dev);

static const struct pci_device_id __devinitdata brcm_ohci_pci_ids [] = 
{ 
        {

    /* handle any USB OHCI controller */
    class:      ((PCI_CLASS_SERIAL_USB << 8) | 0x10),
    class_mask:     ~0,

    /* no matter who makes it */
    vendor:     PCI_ANY_ID,
    device:     PCI_ANY_ID,
    subvendor:  PCI_ANY_ID,
    subdevice:  PCI_ANY_ID,

    }, { /* end: all zeroes */ }
};

struct pci_dev *Pci_dev[2];

/* --------------------------------------------------------------------------
    Name: brcm_ohcd_init
 -------------------------------------------------------------------------- */

static int brcm_ohcd_init (void)
{
    int status;
    
    // Init BRCM USB setup registers for board and HC specific issues
    writel( BRCM_USB_SETUP_REG_VAL, BrcmUsbSetup );
    writel( BRCM_PLL_CONTROL_REG_VAL, BrcmPllControl );

//#ifdef CONFIG_MIPS_BCM7320
//     {
//        setup = (__u32 *) HC2_BASE_ADDR;
//        writel( BRCM_USB_SETUP_REG_VAL, &setup[BrcmUsbSetup] );
//     }   
//#endif

//#ifdef CONFIG_MIPS_BCM7110
//    {   //set 2nd external port to be USB device as per board
//        __u8 *tm_usb_sel = (__u8 *) (TM_TOP_ADR_BASE + 2);
//        writeb( (readb( tm_usb_sel ) & ~0x01), tm_usb_sel );
//    }
//#endif

    //record driver name for resource allocation purposes
    ohci_pci_driver.name = (char *) kmalloc( 30, GFP_KERNEL );
    strcpy( ohci_pci_driver.name, "brcm-usb-ohci" );
    ohci_pci_driver.id_table =  &brcm_ohci_pci_ids [0];

    Pci_dev[0] = kmalloc( sizeof( struct pci_dev ), GFP_KERNEL );
    if( !Pci_dev[0] )
        return -ENOMEM;
    strcpy( Pci_dev[0]->name, "BRCM-OHCI" );
    strcpy( Pci_dev[0]->slot_name, "OHCI-Direct" );
    pci_resource_start(Pci_dev[0], 0) = HC_BASE_ADDR;
    pci_resource_end(Pci_dev[0], 0) = HC_END_ADDR;
    Pci_dev[0]->irq = HC_INT_VECTOR;
    Pci_dev[0]->bus = kmalloc( sizeof(struct pci_bus), GFP_KERNEL );
    if( !Pci_dev[0]->bus )
      return -ENOMEM;

    Pci_dev[0]->bus->number = 0;

    status = ohci_pci_probe (Pci_dev[0], &ohci_pci_driver.id_table[0]);

#ifdef HC2_BASE_ADDR   
      Pci_dev[1] = kmalloc( sizeof( struct pci_dev ), GFP_KERNEL );
      if( !Pci_dev[1] )
          return -ENOMEM;
      strcpy( Pci_dev[1]->name, "BRCM-OHCI2" );
      strcpy( Pci_dev[1]->slot_name, "OHCI2-Direct" );
      pci_resource_start(Pci_dev[1], 0) = HC2_BASE_ADDR;
      pci_resource_end(Pci_dev[1], 0) = HC2_END_ADDR;
      Pci_dev[1]->irq = HC2_INT_VECTOR;
      status = ohci_pci_probe (Pci_dev[1], &ohci_pci_driver.id_table[0]);
#endif

    return( status );
}

/* --------------------------------------------------------------------------
    Name: brcm_ohcd_cleanup
 -------------------------------------------------------------------------- */

static void brcm_ohcd_cleanup (void)
{
    ohci_pci_remove( Pci_dev[0] );
   
    if( Pci_dev[0] )
      {
        if( Pci_dev[0]->bus )
          kfree( Pci_dev[0]->bus );
        kfree( Pci_dev[0] );
      }
    kfree( ohci_pci_driver.name );
#ifdef HC2_BASE_ADDR    
    ohci_pci_remove( Pci_dev[1] );
    if( Pci_dev[1] )
        kfree( Pci_dev[1] );
#endif
}
#endif  //EHCI_BRCM, OHCI_BRCM


//some PCI prototypes
struct pci_pool * pci_pool_create (const char *, struct pci_dev *,
                                   size_t, size_t, size_t, int);
void pci_pool_destroy (struct pci_pool *);
void *pci_pool_alloc (struct pci_pool *, int, dma_addr_t *);
void pci_pool_free (struct pci_pool *, void *, dma_addr_t);

//***********************************************************************
// PCI stub routines for non-PCI based Host Controllers

#define pci_register_driver             brcm_pci_register_driver
static inline int brcm_pci_register_driver(struct pci_driver *drv)
{
#ifdef EHCI_BRCM
    int ret;
#ifdef CONFIG_TIVO_UMA2C //for BCM7317 only
    unsigned int chip_id = (*BCM_CHIPID >> 16);
    if(chip_id == 0x7315)
        return -1;
#endif    
    if ((ret = brcm_ehcd_init()) < 0)
        return ret;
#elif defined(OHCI_BRCM)
    int ret;
    if ((ret = brcm_ohcd_init()) < 0)
        return ret;
#endif  //EHCI_BRCM, OHCI_BRCM

    info(DRIVER_MODS);
        
    return 1;
}

#define pci_unregister_driver           brcm_pci_unregister_driver
static inline void brcm_pci_unregister_driver(struct pci_driver *drv)
{
#ifdef EHCI_BRCM
#ifdef CONFIG_TIVO_UMA2C  //for BCM7317 only
    unsigned int chip_id = (*BCM_CHIPID >> 16);
    if(chip_id != 0x7315)
#endif    
        brcm_ehcd_cleanup();
#elif defined(OHCI_BRCM)
    brcm_ohcd_cleanup();
#endif  //EHCI_BRCM, OHCI_BRCM
}

#define pci_module_init         brcm_pci_module_init
static inline int brcm_pci_module_init(struct pci_driver *drv)
{
    int rc = pci_register_driver (drv);
    if (rc > 0)
        return 0;
    
    if (rc == 0)
        rc = -ENODEV;

    /* if we get here, we need to clean up pci driver instance
     * and return some sort of error */
    pci_unregister_driver (drv);

    return rc;
}

#define pci_set_master                          brcm_pci_set_master
static inline void brcm_pci_set_master(struct pci_dev *dev)
{
}

#define pci_enable_device                       brcm_pci_enable_device
static inline int brcm_pci_enable_device(struct pci_dev *dev)
{
        return 0;
}

#define pci_disable_device                      brcm_pci_disable_device
static inline void brcm_pci_disable_device(struct pci_dev *dev)
{
}

#define pci_find_slot                           brcm_pci_find_slot
static inline struct pci_dev *brcm_pci_find_slot(unsigned int bus, unsigned int devfn)
{
        return NULL;
}

#define pci_read_config_byte            brcm_pci_read_config_byte
static inline int brcm_pci_read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
        *val = 0; 

        return 0;
}

#define pci_write_config_byte           brcm_pci_write_config_byte
static inline int brcm_pci_write_config_byte(struct pci_dev *dev, int where, u8 val)
{
        return 0;
}

#ifdef CONFIG_PM

#define pci_read_config_word            brcm_pci_read_config_word
static inline int brcm_pci_read_config_word(struct pci_dev *dev, int where, u16 *val)
{
        *val = 0; 

        return 0;
}

#define pci_write_config_word           brcm_pci_write_config_word
static inline int brcm_pci_write_config_word(struct pci_dev *dev, int where, u16 val)
{
        return 0;
}

#endif

#define pci_read_config_dword           brcm_pci_read_config_dword
static inline int brcm_pci_read_config_dword(struct pci_dev *dev, int where, u32 *val)
{
        *val = 0; 

        return 0;
}

#define pci_write_config_dword          brcm_pci_write_config_dword
static inline int brcm_pci_write_config_dword(struct pci_dev *dev, int where, u32 val)
{
        return 0;
}

#define pci_set_mwi                                     brcm_pci_set_mwi
static inline void brcm_pci_set_mwi(struct pci_dev *dev)
{
}

#define pci_clear_mwi                           brcm_pci_clear_mwi
static inline void brcm_pci_clear_mwi(struct pci_dev *dev)
{
}

#define pci_set_dma_mask                        brcm_pci_set_dma_mask
static inline int brcm_pci_set_dma_mask(struct pci_dev *dev, u64 mask)
{
        return 0;
}


#endif //__BRCMUSB_H__

#endif // CONFIG_HAS_BRCM_USB
