/*
 * Register defs for Broadcom BCM7020 host controller
 *
 * Copyright (C) 2000 TiVo, Inc. All Rights Reserved.
 */
#ifndef __ASM_BCM7020_H
#define __ASM_BCM7020_H

#include <asm/addrspace.h>

#define CPU_IRQ_SW1		0	/* CPU CP0_STATUS.IP[0] - Software interrupt	*/
#define CPU_IRQ_SW2		1	/* CPU CP0_STATUS.IP[1] - Software interrupt	*/
#define CPU_IRQ_INT0		2	/* CPU CP0_STATUS.IP[2]				*/
#define CPU_IRQ_INT1		3	/* CPU CP0_STATUS.IP[3]				*/
#define CPU_IRQ_INT2		4	/* CPU CP0_STATUS.IP[4]				*/
#define CPU_IRQ_INT3		5	/* CPU CP0_STATUS.IP[5]				*/
#define CPU_IRQ_INT4		6	/* CPU CP0_STATUS.IP[6]				*/
#define CPU_IRQ_INT5		7	/* CPU CP0_STATUS.IP[7] - Cycle Timer interrupt	*/

#define BCM_IRQ_MPEG_INT	8
#define BCM_IRQ_ADP_X_INTR	9
#define BCM_IRQ_IODMA_DONE	10
#define BCM_IRQ_DMA_CMD_DONE	11
#define BCM_IRQ_AUDIO_INTR	12
#define BCM_IRQ_DISP_VSYNC	13
#define BCM_IRQ_XPT_MSG_INTR	14
#define BCM_IRQ_ANA_VSYNC	15
#define BCM_IRQ_DISP_END	16
#define BCM_IRQ_RASTER_TF_CMP	17
#define BCM_IRQ_RASTER_BF_CMP	18
#define BCM_IRQ_CAP_TF_DONE	19
#define BCM_IRQ_CAP_BF_DONE	20
#define BCM_IRQ_TMR_OVFL	21
#define BCM_IRQ_TMR_COMP	22
#define BCM_IRQ_ILL_INT		23
#define BCM_IRQ_SOFT_INT1	24
#define BCM_IRQ_XPT_OVFL_INTR	25
#define BCM_IRQ_ADP_INT		26
#define BCM_IRQ_DMA_CMD_DONE1	27	
#define BCM_IRQ_RESERVED_20	28
#define BCM_IRQ_DISP_VSYNC1	29
#define BCM_IRQ_SMART_CARD0	30
#define BCM_IRQ_SMART_CARD1	31
#define BCM_IRQ_DISP_END1	32
#define BCM_IRQ_RASTER_TF_CMP1	33
#define BCM_IRQ_RASTER_BF_CMP1	34
#define BCM_IRQ_CAP_TF_DONE1	35
#define BCM_IRQ_CAP_BF_DONE1	36
#define BCM_IRQ_TMR_OVFL1	37
#define BCM_IRQ_XPT_INT		38
#define BCM_IRQ_PCI_INT		39

#define BCM_IRQ_PCI_A0		40
#define BCM_IRQ_PCI_A1		41
#define BCM_IRQ_PCI_A2		42
#define BCM_IRQ_PCI_A3		43
#define BCM_IRQ_PCI_DMAPW	44
#define BCM_IRQ_PCI_DMAMW	45
#define BCM_IRQ_PCI_DMAD	46
#define BCM_IRQ_PCI_DMADE	47
#define BCM_IRQ_PCI_TPE		48
#define BCM_IRQ_PCI_PERR	49
#define BCM_IRQ_PCI_TABT	50
#define BCM_IRQ_PCI_MABT	51
#define BCM_IRQ_PCI_SERR	52
#define BCM_IRQ_PCI_APE		53
// 54-55 reserved on RBx, 50-55 reserved on prior

// ISA IRQs only available on PX
#define ALI_IRQ_ISA_0		56
// ...
#define ALI_IRQ_ISA_15		71
// 56-135 are also MS IRQs, but only on P0 and above

// Note: if you increase this, also increase NR_IRQS in <asm/irq.h>
#define BCM_NR_IRQS		136

#ifndef _LANGUAGE_ASSEMBLY
#ifdef __KERNEL__
#include <linux/types.h>

#define __preg8		(volatile u8*)
#define __preg16	(volatile u16*)
#define __preg32	(volatile u32*)
#else /* !__KERNEL__ */
#define __preg8		(volatile unsigned char*)
#define __preg16	(volatile unsigned short*)
#define __preg32	(volatile unsigned int*)
#endif /* __KERNEL__ */
#else /* _LANGUAGE_ASSEMBLY */
#define __preg8
#define __preg16
#define __preg32
#endif /* !_LANGUAGE_ASSEMBLY */

#define BCM_SDRAM_PHYS		0x00000000
#define BCM_SDRAM_BASE 		(KSEG1 + BCM_SDRAM_PHYS)
#define BCM_SDRAM_SIZE		0x08000000
#define BCM_REG_PHYS		0x10000000
#define BCM_REG_BASE 		(KSEG1 + BCM_REG_PHYS)
#define BCM_PCIIO_PHYS		0x13000000
#define BCM_PCIIO_BASE 		(KSEG1 + BCM_PCIIO_PHYS)
#define BCM_PCIIO_SIZE		0x00200000
#define BCM_PCICFG2_PHYS	0x13e00000
#define BCM_PCICFG2_BASE	(KSEG1 + BCM_PCICFG2_PHYS)
#define BCM_PCIMEM_PHYS		0x14000000
#define BCM_PCIMEM_BASE 	(KSEG1 + BCM_PCIMEM_PHYS)
#define BCM_PCIMEMCACHED_BASE	(KSEG0 + BCM_PCIMEM_PHYS)
#define BCM_PCIMEM_SIZE		0x04000000
#define BCM_PCICFG_PHYS		0x1bfa0000
#define BCM_PCICFG_BASE 	(KSEG1 + BCM_PCICFG_PHYS)
#define	BCM_LBUS_PHYS		0x1c000000


#define __SYSREG(offset)	__preg32(BCM_REG_BASE + 0 + offset)

#define BCM_MEM_BASE		__SYSREG(0x000)		/* Memory Base Address Register (R/W) */
#define BCM_MEM_CTRL		__SYSREG(0x008)		/* Memory Control Register (R/W) */
#define BCM_GP_CTRL		__SYSREG(0x00c)
#define BCM_CPU_CTRL		__SYSREG(0x010)		/* CPU Control Register (R/W) */
#define BCM_INTR_ENA		__SYSREG(0x018)		/* Interrupt Enable Register (W) */
#define BCM_INTR_CLEAR		__SYSREG(0x01c)		/* Interrupt Clear Register (W) */
#define BCM_INTR_STATUS		__SYSREG(0x01c)		/* Interrupt Status Register (R) */
#define BCM_TIMER_CMP		__SYSREG(0x020)		/* Timer Counter Compare Register (R/W) */
#define BCM_TIMER_COUNT		__SYSREG(0x024)		/* Timer Counter Read/Load Register (R/W) */
#define BCM_MAIL_BOX1		__SYSREG(0x028)		/* CPU/Paxel Mail Box 1 Register (R/W) */
#define BCM_MAIL_BOX2		__SYSREG(0x02c)		/* CPU/Paxel Mail Box 2 Register (R/W) */
#define BCM_DEV_ID		__SYSREG(0x0fc)		/* Device ID Register (R) */
#define BCM_IO_CS0_ADDR		__SYSREG(0x200)		/* IO Chip Select 0 Address Range Register (R/W) */
#define BCM_IO_CS1_ADDR		__SYSREG(0x204)		/* IO Chip Select 1 Address Range Register (R/W) */
#define BCM_IO_CS2_ADDR		__SYSREG(0x208)		/* IO Chip Select 2 Address Range Register (R/W) */
#define BCM_IO_CS3_ADDR		__SYSREG(0x20c)		/* IO Chip Select 3 Address Range Register (R/W) */
#define BCM_IO_CS4_ADDR		__SYSREG(0x210)		/* IO Chip Select 4 Address Range Register (R/W) */
#define BCM_IO_CS5_ADDR		__SYSREG(0x214)		/* IO Chip Select 5 Address Range Register (R/W) */
#define BCM_IO_NULL_TERM	__SYSREG(0x230)		/* IO Termination States Register (R/W) */
#define BCM_IO_DMA0_REC		__SYSREG(0x240)		/* IO DMA0 Record Starting Address Register (R/W) */
#define BCM_IO_DMA0_CTL		__SYSREG(0x244)		/* IO DMA0 Control Register (R/W) */
#define BCM_IO_DMA0_DEB		__SYSREG(0x248)		/* IO DMA0 Debug Register (R/W) */
#define BCM_IO_DMA1_REC		__SYSREG(0x250)		/* IO DMA1 Record Starting Address Register (R/W) */
#define BCM_IO_DMA1_CTL		__SYSREG(0x254)		/* IO DMA1 Control Register (R/W) */
#define BCM_IO_DMA1_DEB		__SYSREG(0x258)		/* IO DMA1 Debug Register (R/W) */
#define BCM_PCI_INTR_REG	__SYSREG(0x300)		/* PCI Interrupt Register (R/W) */
#define BCM_PCI_INTR_STAT	__SYSREG(0x304)		/* RBx PCI Interrupt Status Register (R/W) */
#define BCM_PCI_BUS_NUM		__SYSREG(0x304)		/* PCI Bus Number Register (R/W) */
#define BCM_PCI_DMA_CTRL	__SYSREG(0x308)		/* PCI DMA Controller Register (R/W) */
#define BCM_PCI_DMA_CURRENT_DESC __SYSREG(0x30c) 	/* PCI DMA Current Descriptor Register (R/W) */
#define BCM_PCI_DMA_DESC_WRD0	__SYSREG(0x310)		/* PCI DMA Current Descriptor Word 0 Register (R/W) */
#define BCM_PCI_DMA_DESC_WRD1	__SYSREG(0x314)		/* PCI DMA Current Descriptor Word 1 Register (R/W) */
#define BCM_PCI_DMA_DESC_WRD2	__SYSREG(0x318)		/* PCI DMA Current Descriptor Word 2 Register (R/W) */
#define BCM_PCI_DMA_CUR_MEM_BYTE_CNT __SYSREG(0x31c)	/* PCI DMA Current Memory Byte Count Register (R/W) */


#define __VIDREG(offset)	__preg32(BCM_REG_BASE + 0x2000 + 4 * (offset))

#define BCM_RB_TMC_CNTRL_CONFIG_REG	__VIDREG(0x00)
#define BCM_RB_TMC_MODE_CONTROL_REG	__VIDREG(0x01)
#define BCM_RB_TMC_DELAY_REG		__VIDREG(0x1e)
#define BCM_RB_SOFT_RESET_WITHDRAW_REG1	__VIDREG(0x4+0x28)


#define __CFGREG8(offset)	__preg8(BCM_PCICFG_BASE + 0 + offset)
#define __CFGREG32(offset)	__preg32(BCM_PCICFG_BASE + 0 + offset)
#define __CFG2REG8(offset)	__preg8(BCM_PCICFG2_BASE + 0 + offset)
#define __CFG2REG32(offset)	__preg32(BCM_PCICFG2_BASE + 0 + offset)

#define BCM_PCI_CMD_SELF	__CFGREG32(0x004)
#define BCM_PCI_LAT_SELF	__CFGREG32(0x00c)
#define BCM_PCI_IO_BASE		__CFGREG32(0x010)	/* PCI BAR0 */
#define BCM_PCI_MEM_BASE	__CFGREG32(0x014)	/* PCI BAR1 */
#define BCM_PCI_REG_BASE	__CFGREG32(0x018)	/* PCI BAR2 */
#define BCM_PCI_PCU_BASE	__CFGREG32(0x01c)	/* PCI BAR3 */
#define BCM_PCI_MEM_REMAP	__CFGREG32(0x050)
#define BCM_PCI_IO_REMAP	__CFGREG32(0x054)
#define BCM_PCI_MEM_BASE_ADDR	__CFGREG32(0x058)
#define BCM_PCI_MEM_SIZE	__CFGREG32(0x05c)
#define BCM_PCI_MYSTERY_REG	__CFGREG32(0x060)
#define  BCM_PCI_MR_SWAP	0x02	/* bit1: byte swap */
#define  BCM_PCI_MR_ARB		0x04	/* bit2: external arbiter enable */
#define BCM_PCI_CF8		__CFG2REG32(0xcf8)	/* PCI config space address (R?/W) */
#define BCM_PCI_CFC		__CFG2REG32(0xcfc)	/* PCI config space data (R/W) */

///////////////////////////////////////////////////////////////////////////////
//
// CACHES
//
///////////////////////////////////////////////////////////////////////////////

#define ICACHE_SIZE                     0x4000
#define ICACHE_LINESIZE                 0x0010
#define DCACHE_SIZE                     0x4000
#define DCACHE_LINESIZE                 0x0010

#endif /* __ASM_BCM7020_H */
