#ifndef _ASM_IRQ_TIVO_H
#define _ASM_IRQ_TIVO_H

#define NR_IRQS	160

#include <asm/io.h>

#ifndef NEW_MS_BASE_ADDR
#define NEW_MS_BASE_ADDR	0xf9800000
#endif
#ifndef OLD_MS_BASE_ADDR
#define OLD_MS_BASE_ADDR	0xffd00000
#endif

extern unsigned int ms_base_addr;
#define MS_BASE_ADDR (ms_base_addr)

extern void disable_irq(unsigned int);
extern void enable_irq(unsigned int);
static __inline__ int irq_cannonicalize(int irq)
{
	return irq;
}

/*
 * Mediaswitch interrupt register addresses
 * Can only be accessed after TLB
 * intialized (or in real mode).
 */
#define MEDIASWITCH_IER		(MS_BASE_ADDR+0x000c)
#define MEDIASWITCH_ISR		(MS_BASE_ADDR+0x0010)
#define MEDIASWITCH_IBASE	32

#define MEDIASWITCH_EIER	(MS_BASE_ADDR+0x0014)
#define MEDIASWITCH_EISR	(MS_BASE_ADDR+0x0018)
#define MEDIASWITCH_EIBASE	64

#define MEDIASWITCH2_IBASE	96
#define MEDIASWITCH2_EIBASE	128

#ifdef COMBO_BOX
/* TL750 addresses */
/* These only apply to combo box.*/
#define TL750_SDRAM	0xf8000000	/* TL750 attached SDRAM (CS0) 16MB */
#define TL750_REGISTER	0xf9000000	/* TL750 internal register (CS1) 8MB */
#endif

/*
 * Mediaswitch 2.  
 *
 * Mediaswitch2 is new to the combo box hardware.  Initially the address 
 * space is shared with the mediawitch.  The define for M2ADDRESSCOMPAT 
 * controls whether the address map reflects the new MS2 addresses or the
 * shared chip select with the MS.
 */

/* General Registers */
#define MEDIASWITCH2_SR		(MS_BASE_ADDR+0x1000)	/* System Register */
#define MEDIASWITCH2_CR		(MS_BASE_ADDR+0x1004)	/* Control Register */
#define MEDIASWITCH2_SR2	(MS_BASE_ADDR+0x1008)	/* Status Register */
#define MEDIASWITCH2_IER	(MS_BASE_ADDR+0x100c)	/* Interrupt Mask Register */
#define MEDIASWITCH2_ISR	(MS_BASE_ADDR+0x1010)	/* Interrupt Status Register */
#define MEDIASWITCH2_EIER	(MS_BASE_ADDR+0x1014)	/* Hardware Error Interrupt Mask Register */
#define MEDIASWITCH2_EISR	(MS_BASE_ADDR+0x1018)	/* Hardware Error Interrupt Status Register */
#define MEDIASWITCH2_VSBR	(MS_BASE_ADDR+0x1020)	/* Viedo Slice Boundary Register */
#define MEDIASWITCH2_I2C	(MS_BASE_ADDR+0x1100)	/* I2C Data/Control Register */

/* DMA Channel Registers -- DMA Address/Count */
#define MEDIASWITCH2_DMA0VIADDR	(MS_BASE_ADDR+0x1200)	/* DMA Channel 0, Video Input DMA Address */
#define MEDIASWITCH2_DMA0VICNT	(MS_BASE_ADDR+0x1204)	/* DMA Channel 0, Video Input Next Byte Count */
#define MEDIASWITCH2_DMA1AIADDR	(MS_BASE_ADDR+0x1208)	/* DMA Channel 1, Audio Input DMA Address */
#define MEDIASWITCH2_DMA1AICNT	(MS_BASE_ADDR+0x120c)	/* DMA Channel 1, Audio Input Next Byte Count */

/* The next two are unused. */
#define MEDIASWITCH2_DMA2XXXXX	(MS_BASE_ADDR+0x1210)	/* DMA Channel 2 */
#define MEDIASWITCH2_DMA2YYYYY	(MS_BASE_ADDR+0x1214)	/* DMA Channel 2 */

#define MEDIASWITCH2_DMA3EADDR	(MS_BASE_ADDR+0x1218)	/* DMA Channel 3, Event DMA Address */
#define MEDIASWITCH2_DMA3ECNT	(MS_BASE_ADDR+0x121c)	/* DMA Channel 3, Event Next Byte Count */
#define MEDIASWITCH2_DMA4OIADDR	(MS_BASE_ADDR+0x1220)	/* DMA Channel 4, OSLink Input DMA Address */
#define MEDIASWITCH2_DMA4OICNT	(MS_BASE_ADDR+0x1224)	/* DMA Channel 4, OSLink Input Next Byte Count */
#define MEDIASWITCH2_DMA5OOADDR	(MS_BASE_ADDR+0x1228)	/* DMA Channel 5, OSLink Output DMA Address */
#define MEDIASWITCH2_DMA5OOCNT	(MS_BASE_ADDR+0x122c)	/* DMA Channel 5, OSLink Output Next Byte Count */

/* The next four are unused */
#define MEDIASWITCH2_DMA6AOADDR	(MS_BASE_ADDR+0x1230)	/* DMA Channel 6, Audio Output DMA Address */
#define MEDIASWITCH2_DMA6AOCNT	(MS_BASE_ADDR+0x1234)	/* DMA Channel 6, Audio Output Next Byte Count */
#define MEDIASWITCH2_DMA7OADDR	(MS_BASE_ADDR+0x1238)	/* DMA Channel 7, OSD Output DMA Address */
#define MEDIASWITCH2_DMA7OCNT	(MS_BASE_ADDR+0x123c)	/* DMA Channel 7, OSD Output Next Byte Count */

/* DMA Channels -- Current Addr/Current Byte Count */
#define MEDIASWITCH2_DMA0VICUR	(MS_BASE_ADDR+0x1240)	/* DMA Channel 0, Video Input Current Address */
#define MEDIASWITCH2_DMA0VICBC	(MS_BASE_ADDR+0x1244)	/* DMA Channel 0, Video Input Current Byte Count */
#define MEDIASWITCH2_DMA1AICUR	(MS_BASE_ADDR+0x1248)	/* DMA Channel 1, Audio Input Current Address */
#define MEDIASWITCH2_DMA1AICBC	(MS_BASE_ADDR+0x124c)	/* DMA Channel 1, Audio Input Next Byte Count */

/* The two are unused */
#define MEDIASWITCH2_DMA2VBCUR	(MS_BASE_ADDR+0x1250)	/* DMA Channel 2, VBI Current Address */
#define MEDIASWITCH2_DMA2VBCBC	(MS_BASE_ADDR+0x1254)	/* DMA Channel 2, VBI Current Byte Count */

#define MEDIASWITCH2_DMA3ECUR	(MS_BASE_ADDR+0x1258)	/* DMA Channel 3, Event Current Address */
#define MEDIASWITCH2_DMA3ECBC	(MS_BASE_ADDR+0x125c)	/* DMA Channel 3, Event Current Byte Count */
#define MEDIASWITCH2_DMA4OICUR	(MS_BASE_ADDR+0x1260)	/* DMA Channel 4, OSLink Input Current Address */
#define MEDIASWITCH2_DMA4OICBC	(MS_BASE_ADDR+0x1264)	/* DMA Channel 4, OSLink Input Current Byte Count */
#define MEDIASWITCH2_DMA5OOCUR	(MS_BASE_ADDR+0x1268)	/* DMA Channel 5, OSLink Output Current Address */
#define MEDIASWITCH2_DMA5OOCBC	(MS_BASE_ADDR+0x126c)	/* DMA Channel 5, OSLink Output Current Byte Count */

/* Next four not used */
#define MEDIASWITCH2_DMA6AOCUR	(MS_BASE_ADDR+0x1270)	/* DMA Channel 6, Audio Output Current Address */
#define MEDIASWITCH2_DMA6AOCBC	(MS_BASE_ADDR+0x1274)	/* DMA Channel 6, Audio Output Current Byte Count */
#define MEDIASWITCH2_DMA7OCUR	(MS_BASE_ADDR+0x1278)	/* DMA Channel 7, OSD Output Current Address */
#define MEDIASWITCH2_DMA7OCBC	(MS_BASE_ADDR+0x127c)	/* DMA Channel 7, OSD Output Current Byte Count */

/* ST55xx Registers */
#define MEDIASWITCH2_ST2HIMR	(MS_BASE_ADDR+0x1400)	/* ST55xx to Host Interrupt Mask Register */
#define MEDIASWITCH2_ST2HIR	(MS_BASE_ADDR+0x1404)	/* ST55xx to Host Interrupt Register */
#define MEDIASWITCH2_ST2HCR0	(MS_BASE_ADDR+0x1408)	/* ST55xx to Host Command Register 0 */
#define MEDIASWITCH2_ST2HCR1	(MS_BASE_ADDR+0x140c)	/* ST55xx Host Command Register 1 */
#define MEDIASWITCH2_H2STCR0	(MS_BASE_ADDR+0x1410)	/* Host to ST55xx Command Register 0 */
#define MEDIASWITCH2_H2STCR1	(MS_BASE_ADDR+0x1414)	/* Host to ST55xx Command Register 1 */
#define MEDIASWITCH2_H2STIR	(MS_BASE_ADDR+0x1418)	/* Host to ST55xx Interrupt Register */
#define MEDIASWITCH2_H2STIMR	(MS_BASE_ADDR+0x141c)	/* Host to ST55xx Interrupt Mask Register */

#define MEDIASWITCH2_SIGREG	(MS_BASE_ADDR+0x1600)	/* Media Switch Signature Register */

#define MEDIASWITCH_REV_MASK    (0xff)
#define MEDIASWITCH2_SIGNATURE	0xabcd1200	/* SIGREG should return this value if the mediaswitch2 is present */
#define MEDIASWITCH3_SIGNATURE	0xabcd5600	/* SIGREG should return this value if the mediaswitch2 is present */
#define MEDIASWITCH4_SIGNATURE  0xabcd0100      /* SIGREG should return this value if the mediaswitch4 is present. FIXME, MS4 value is dummy */
#define MEDIASWITCH2_MS1INTR	(1 << 31)


/* ms2probe returns true for either PX or P0+ combo boards, ms3probe only for P0+ */

/* Routine to probe for mediaswitch2 */
extern inline int 
ms2probe(void) {
  int signature = inl(MEDIASWITCH2_SIGREG) & ~MEDIASWITCH_REV_MASK;
  return ((signature == MEDIASWITCH2_SIGNATURE) || (signature == MEDIASWITCH3_SIGNATURE) || (signature == MEDIASWITCH4_SIGNATURE));
}

extern inline int 
ms3probe(void) {
  int signature = inl(MEDIASWITCH2_SIGREG) & ~MEDIASWITCH_REV_MASK;
  return ((signature == MEDIASWITCH3_SIGNATURE) || (signature == MEDIASWITCH4_SIGNATURE));
}

extern inline int
ms4probe(void) {
  int signature = inl(MEDIASWITCH2_SIGREG) & ~MEDIASWITCH_REV_MASK;
  return (signature == MEDIASWITCH4_SIGNATURE);
}

/* This function does not take into account the fact that the last 2 nibbles
 * have a different meaning on MSvn (n < 4)
 * See drivers/common/fpga/tivo1/fpga_hw.c:fpga_get_ms_revision
 */
extern inline int
msrevision(void) {
  return (inl(MEDIASWITCH2_SIGREG) & MEDIASWITCH_REV_MASK);
}

/* Mediaswitch 2 ISR */
#define IRQ_MS2_MS1		(MEDIASWITCH2_IBASE+0)		/* Mediaswitch1 interrupt request */
#define IRQ_MS2_ST2HSI		(MEDIASWITCH2_IBASE+1)		/* ST55xx to host software interrupt */
#define IRQ_MS2_H2STI		(MEDIASWITCH2_IBASE+2)		/* ST55xx to host hardware interrupt */
#define IRQ_MS2_MPEG2		(MEDIASWITCH2_IBASE+3)		/* Placeholder for Icompression encoder */
#define IRQ_MS2_I2C		(MEDIASWITCH2_IBASE+4)		/* I2C interrupt */
#define IRQ_MS2_TL750		(MEDIASWITCH2_IBASE+5)		/* TL750 graphics chip */
#define IRQ_MS2_AFRAME		(MEDIASWITCH2_IBASE+8)	/* audio frame event */
#define IRQ_MS2_APES		(MEDIASWITCH2_IBASE+9)	/* audio PES event */
#define IRQ_MS2_PFRAME		(MEDIASWITCH2_IBASE+10)	/* P frame event */
#define IRQ_MS2_BFRAME		(MEDIASWITCH2_IBASE+11)	/* B frame event */
#define IRQ_MS2_IFRAME		(MEDIASWITCH2_IBASE+12)	/* I frame event */
#define IRQ_MS2_GOP		(MEDIASWITCH2_IBASE+13)	/* GOP event */
#define IRQ_MS2_SEQ		(MEDIASWITCH2_IBASE+14)	/* sequence event */
#define IRQ_MS2_VPES		(MEDIASWITCH2_IBASE+15)	/* video PES event */
#define IRQ_MS2_OSLO_NEXT	(MEDIASWITCH2_IBASE+18)	/* OS Link  out DMA next empty */
#define IRQ_MS2_OSLI_NEXT	(MEDIASWITCH2_IBASE+19)	/* OS Link input  DMA next empty */
#define IRQ_MS2_EVENTDMA_NEXT	(MEDIASWITCH2_IBASE+20)	/* event DMA next empty */
#define IRQ_MS2_AIDMA_NEXT	(MEDIASWITCH2_IBASE+22)	/* audio input DMA next empty */
#define IRQ_MS2_VIDMA_NEXT	(MEDIASWITCH2_IBASE+23)	/* video input DMA next empty */
#define IRQ_MS2_OSLO_IDLE	(MEDIASWITCH2_IBASE+26)	/* OS Link Output Idle */
#define IRQ_MS2_OSLI_IDLE	(MEDIASWITCH2_IBASE+27)	/* OS Link Input Idle */
#define IRQ_MS2_EVENTDMA_IDLE	(MEDIASWITCH2_IBASE+28)	/* event DMA Idle */
#define IRQ_MS2_AIDMA_IDLE	(MEDIASWITCH2_IBASE+30)	/* audio input DMA next empty */
#define IRQ_MS2_VIDMA_IDLE	(MEDIASWITCH2_IBASE+31)	/* video input DMA next empty */

/*
 * Mediaswitch error interrupt numbers.
 */
#define IRQ_MS2E_I2CARB		(MEDIASWITCH2_EIBASE+20)	/* I2C arbitration error */
#define IRQ_MS2E_I2CACK		(MEDIASWITCH2_EIBASE+21)	/* I2C acknowledge error */
#define IRQ_MS2E_OSLOUNDER	(MEDIASWITCH2_EIBASE+28)	/* OS Link Output FIFO Underrun */
#define IRQ_MS2E_OSLIOVER	(MEDIASWITCH2_EIBASE+29)	/* OS Link Input Overflow */
#define IRQ_MS2E_AIOVER		(MEDIASWITCH2_EIBASE+30)	/* Audio input FIFO overflow error */
#define IRQ_MS2E_VIOVER		(MEDIASWITCH2_EIBASE+31)	/* Video input FIFO overflow error */

/*
 * Basic PPC IRQ Numbers
 */
#define IRQ_CRITICAL		0
#define	IRQ_403SERIAL_RX	4
#define IRQ_403SERIAL_TX	5
#define IRQ_403DMA0		8
#define IRQ_403DMA1		9
#define IRQ_403DMA2		10
#define IRQ_403DMA3		11
#ifdef CONFIG_TCD_EVAL
#define IRQ_8390		27
#define IRQ_UART		28
#else
#define IRQ_MEDIASWITCH		27
#define IRQ_MPEGDECODE		28
#define IRQ_8390		29
#define IRQ_UART		30
#define IRQ_MODEM		31
#endif

/*
 * Mediaswitch normal interrupt numbers.
 * Bit numbers in the documentation are opposite
 * PPC bit numbering (LSB is zero rather than MSB)
 */
#define IRQ_MS_IDE		(MEDIASWITCH_IBASE+0)	/* ide interrupt. */
#define IRQ_MS_AIIFL1 		(MEDIASWITCH_IBASE+1)	/* analog input FL1 */
#define IRQ_MS_DAALS		(MEDIASWITCH_IBASE+2)	/* DAA Line Status */
#define IRQ_MS_CXD              (MEDIASWITCH_IBASE+3)     /* SONY CXD */
#define IRQ_MS_I2C		(MEDIASWITCH_IBASE+4)	/* I2C ready */
#define IRQ_MS_ATS		(MEDIASWITCH_IBASE+5)	/* audio time stamp */
#define IRQ_MS_VFTS		(MEDIASWITCH_IBASE+6)	/* video frame time stamp */
#define IRQ_MS_VSYNCTS		(MEDIASWITCH_IBASE+7)	/* video sync time stamp */

#define IRQ_MS_AFRAME		(MEDIASWITCH_IBASE+8)	/* audio frame event */
#define IRQ_MS_APES		(MEDIASWITCH_IBASE+9)	/* audio PES event */
#define IRQ_MS_PFRAME		(MEDIASWITCH_IBASE+10)	/* P frame event */
#define IRQ_MS_BFRAME		(MEDIASWITCH_IBASE+11)	/* B frame event */
#define IRQ_MS_IFRAME		(MEDIASWITCH_IBASE+12)	/* I frame event */
#define IRQ_MS_GOP		(MEDIASWITCH_IBASE+13)	/* GOP event */
#define IRQ_MS_SEQ		(MEDIASWITCH_IBASE+14)	/* sequence event */
#define IRQ_MS_VPES		(MEDIASWITCH_IBASE+15)	/* video PES event */

#define IRQ_MS_OSDDMA_NEXT	(MEDIASWITCH_IBASE+16)	/* OSD DMA next empty */
#define IRQ_MS_AODMA_NEXT	(MEDIASWITCH_IBASE+17)	/* audio out DMA next empty */
#define IRQ_MS_VODMA_NEXT	(MEDIASWITCH_IBASE+18)	/* video out DMA next empty */
#define IRQ_MS_IDEDMA_NEXT	(MEDIASWITCH_IBASE+19)	/* IDE DMA next empty */
#define IRQ_MS_EVENTDMA_NEXT	(MEDIASWITCH_IBASE+20)	/* event DMA next empty */
#define IRQ_MS_VBIDMA_NEXT	(MEDIASWITCH_IBASE+21)	/* VBI DMA next empty */
#define IRQ_MS_AIDMA_NEXT	(MEDIASWITCH_IBASE+22)	/* audio input DMA next empty */
#define IRQ_MS_VIDMA_NEXT	(MEDIASWITCH_IBASE+23)	/* video input DMA next empty */

#define IRQ_MS_OSDDMA_IDLE	(MEDIASWITCH_IBASE+24)	/* OSD DMA next empty */
#define IRQ_MS_AODMA_IDLE	(MEDIASWITCH_IBASE+25)	/* audio out DMA next empty */
#define IRQ_MS_VODMA_IDLE	(MEDIASWITCH_IBASE+26)	/* video output DMA idle */
#define IRQ_MS_IDEDMA_IDLE	(MEDIASWITCH_IBASE+27)	/* IDE DMA next empty */
#define IRQ_MS_EVENTDMA_IDLE	(MEDIASWITCH_IBASE+28)	/* event DMA next empty */
#define IRQ_MS_VBIDMA_IDLE	(MEDIASWITCH_IBASE+29)	/* VBI DMA next empty */
#define IRQ_MS_AIDMA_IDLE	(MEDIASWITCH_IBASE+30)	/* audio input DMA next empty */
#define IRQ_MS_VIDMA_IDLE	(MEDIASWITCH_IBASE+31)	/* video input DMA next empty */

/*
 * Mediaswitch error interrupt numbers.
 */
#define IRQ_MSE_I2CARB		(MEDIASWITCH_EIBASE+20)	/* I2C arbitration error */
#define IRQ_MSE_I2CACK		(MEDIASWITCH_EIBASE+21)	/* I2C acknowledge error */
#define IRQ_MSE_IDEPIO		(MEDIASWITCH_EIBASE+22)	/* IDE PIO error */
#define IRQ_MSE_PIOACCESS	(MEDIASWITCH_EIBASE+23)	/* PIO access error */
#define IRQ_MSE_AUNDER		(MEDIASWITCH_EIBASE+27)	/* Audio FIFO underrun error */
#define IRQ_MSE_IDEOVER		(MEDIASWITCH_EIBASE+28)	/* IDE FIFO overflow error */
#define IRQ_MSE_VBIOVER		(MEDIASWITCH_EIBASE+29)	/* VBI FIFO overflow error */
#define IRQ_MSE_AIOVER		(MEDIASWITCH_EIBASE+30)	/* Audio input FIFO overflow error */
#define IRQ_MSE_VIOVER		(MEDIASWITCH_EIBASE+31)	/* Video input FIFO overflow error */

extern unsigned long ms_isr_cleared_g;
extern unsigned long  irq_ms_g;
extern inline void 
clear_current_ms1_isr(void) {
  ms_isr_cleared_g = 1;
  outl((1 << (31 - irq_ms_g)), MEDIASWITCH_ISR);  
}

extern inline void 
clear_current_ms2_isr(void) {
  ms_isr_cleared_g = 1;
  outl((1 << (31 - irq_ms_g)), MEDIASWITCH2_ISR);  
}

#endif
